/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Box
Version:   1.0
ID:        ID_BOX
Status:    Stable
Category:  GUI
Date:      June 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Short:     Draws rectangular graphics.
Keywords:  box, graphics, line, square, rectangle, border, highlight, fill

###DESCRIPTION###
<p>The Box is a simple drawable object used for drawing bevelled boxes and
filled areas inside rendered objects.  By default a Box will take up the
entire width and height of its container, but powerful features are also
provided that allow you to create complicated shapes out of multiple
rectangles.</p>

<p>Here are some examples for creating Box graphics.  The first creates a
raised, bevelled box with grey colouring.</p>

<pre>
   &lt;box colour="#808080" shadow="#303030" highlight="#ffffff" raised/&gt;
</pre>

<p>This example creates a 10x10 box with a black border that is 2 pixels
thick:</p>

<pre>
   &lt;box thickness="2" width="10" height="10" border="#000000"/&gt;
</pre>

<p>If you want to draw multiple boxes with one object, you can use the Boxes
field.  This next example draws a box that fills up the area of its container,
and also draws a smaller box that is offset by 5 pixels from each margin:</p>

<pre>
   &lt;box boxes="(0,0,!0,!0) (5,5,!5,!5)" colour0="255,0,0"
     colour1="255,255,255"/&gt;
</pre>

<p>For more detail on how to use the Box class, please refer to the
documentation for the individual fields.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/functions/strtofloat.c>
//#include <pandora/modules/render.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_BOX ((FLOAT)1.0)

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC,
   VER_BOX, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003. All rights reserved.",
   "June 2003",
   "Box"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR BoxClass = 0;
//static OBJECTPTR modRender = NULL;
//static struct RenderBase *RenderBase;

/****************************************************************************
** Box class.
*/

#define MAX_BOXES 10

struct Boxes {
   FLOAT XCoord;         /* X Coordinate */
   FLOAT YCoord;         /* Y Coordinate */
   FLOAT Width;          /* Width */
   FLOAT Height;         /* Height */
   FLOAT XOffset;        /* X offset */
   FLOAT YOffset;        /* Y offset */
   WORD Dimensions;      /* Dimension flags */
   WORD Round;           /* Size of round corners, in pixels.  Min. 2 */
   WORD BorderOpacity;   /* Opacity 0 - 255 */
   WORD Opacity;         /* Opacity 0 - 255 */
   LONG Flags;
   struct RGB Highlight;
   struct RGB Shadow;
   struct RGB Colour;
};

/*** Public box flags are in lower word ***/
#define BXF_NOFXDRAW      0x00000001
#define BXF_REPEATEFFECT  0x00000002
#define BXF_ACTIVEDRAW    0x00000004
/*** Private flags are in upper word ***/
#define BXF_HIGHLIGHT     0x00010000
#define BXF_SHADOW        0x00020000
#define BXF_COLOUR        0x00040000
#define BXF_FILL          0x00080000
#define BXF_SUNKEN        0x00100000
#define BXF_RAISED        0x00200000
#define BXF_OPACITY       0x00400000
#define BXF_BORDEROPACITY 0x00800000

static struct FieldDef BoxFlags[] = {
   { "NOFXDRAW",     BXF_NOFXDRAW },
   { "REPEATEFFECT", BXF_REPEATEFFECT },
   { "ACTIVEDRAW",   BXF_ACTIVEDRAW },
   { NULL, NULL }
};

struct Box {
   OBJECT_HEADER
   LONG     Thickness;
   LONG     Frame;
   OBJECTID DrawableID;       /* The object that should be rendered to */
   LONG     Visible;
   LONG     TotalBoxes;
   LONG     Effect;        /* Special effect for drawing in real time */
   LONG     Delay;         /* If the effect should be delayed, specify the time in milliseconds */
   LONG     Duration;      /* The required duration of an active effect, expressed in milliseconds */
   LONG     Flags;
   FLOAT    Brightness;    /* Alter the brightness value of colours by this value (default 100%) */
   FLOAT    DarkFocus;     /* Percentage to lower the colours by if the parent drawable does not have the focus */

   PRIVATE_FIELDS
   struct RGB Highlight;
   struct RGB Shadow;
   struct RGB Colour;
   struct Boxes Boxes[MAX_BOXES];
   struct BitmapPen *Pen;
   WORD   *Frames;
   WORD   TotalFrames;
   WORD   EffectActive;
   FLOAT  Ticks;
   WORD   CurrentFrame;
   BYTE   Tag;
};

#define BFX_FADE    1
#define BFX_FADEOUT 2

static struct FieldDef BoxEffects[] = {
   { "FADE",    BFX_FADE },
   { "FADEIN",  BFX_FADE },
   { "FADEOUT", BFX_FADEOUT },
   { NULL, NULL }
};

/*** Field Prototypes ***/

static ERROR GET_BorderOpacity(struct Box *, FLOAT *);
static ERROR GET_ColourRGB(struct Box *, struct RGB *);
static ERROR GET_Fill(struct Box *, LONG *);
static ERROR GET_Height(struct Box *, struct Variable *);
static ERROR GET_HighlightRGB(struct Box *, struct RGB *);
static ERROR GET_Opacity(struct Box *, FLOAT *);
static ERROR GET_ShadowRGB(struct Box *, struct RGB *);
static ERROR GET_Width(struct Box *, struct Variable *);
static ERROR GET_X(struct Box *, struct Variable *);
static ERROR GET_XOffset(struct Box *, struct Variable *);
static ERROR GET_Y(struct Box *, struct Variable *);
static ERROR GET_YOffset(struct Box *, struct Variable *);

static ERROR SET_Border(struct Box *, STRING);
static ERROR SET_BorderOpacity(struct Box *, FLOAT *);
static ERROR SET_Boxes(struct Box *, STRING);
static ERROR SET_Brightness(struct Box *, FLOAT *);
static ERROR SET_Colour(struct Box *, STRING);
static ERROR SET_Fill(struct Box *, LONG *);
static ERROR SET_Flags(struct Box *, LONG *);
static ERROR SET_Frames(struct Box *, STRING);
static ERROR SET_Height(struct Box *, struct Variable *);
static ERROR SET_Highlight(struct Box *, STRING);
static ERROR SET_Opacity(struct Box *, FLOAT *);
static ERROR SET_Raised(struct Box *, LONG *);
static ERROR SET_Round(struct Box *, LONG *);
static ERROR SET_Shadow(struct Box *, STRING);
static ERROR SET_Sunken(struct Box *, LONG *);
static ERROR SET_Thickness(struct Box *, LONG *);
static ERROR SET_TotalBoxes(struct Box *, LONG *);
static ERROR SET_Width(struct Box *, struct Variable *);
static ERROR SET_X(struct Box *, struct Variable *);
static ERROR SET_XOffset(struct Box *, struct Variable *);
static ERROR SET_Y(struct Box *, struct Variable *);
static ERROR SET_YOffset(struct Box *, struct Variable *);

/*** Class Definition ***/

static struct FieldArray BoxFields[] = {
   { "Thickness",  0, FDF_LONG|FDF_RW,      0,         NULL, SET_Thickness },
   { "Frame",      0, FDF_LONG|FDF_RW,      0,         NULL, NULL },
   { "Drawable",   0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Visible",    0, FDF_LONG|FDF_RW,      0,         NULL, NULL },
   { "TotalBoxes", 0, FDF_LONG|FDF_RW,      0,         NULL, SET_TotalBoxes },
   { "Effect",     0, FDF_LONG|FDF_LOOKUP|FDF_RI, (LONG)&BoxEffects, NULL, NULL },
   { "Delay",      0, FDF_LONG|FDF_RI,      0,        NULL, NULL },
   { "Duration",   0, FDF_LONG|FDF_RW,      0,        NULL, NULL },
   { "Flags",      0, FDF_LONGFLAGS|FDF_RW, (LONG)&BoxFlags, NULL, SET_Flags },
   { "Brightness", 0, FDF_FLOAT|FDF_RW,     0,        NULL, SET_Brightness },
   { "DarkFocus",  0, FDF_FLOAT|FDF_RI,     0,        NULL, NULL },
   /*** Virtual Fields ***/
   { "Border",       0, FDF_STRING|FDF_W,    0, NULL,             SET_Border },
   { "BorderOpacity",0, FDF_FLOAT|FDF_RW,    0, GET_BorderOpacity,SET_BorderOpacity },
   { "Boxes",        0, FDF_STRING|FDF_W,    0, NULL,             SET_Boxes  },
   { "Colour",       0, FDF_STRING|FDF_W,    0, NULL,             SET_Colour },
   { "ColourRGB",    0, FDF_POINTER|FDF_R,   0, GET_ColourRGB,    NULL },
   { "Frames",       0, FDF_STRING|FDF_W,    0, NULL,             SET_Frames },
   { "Fill",         0, FDF_LONG|FDF_RW,     0, GET_Fill,         SET_Fill },
   { "Highlight",    0, FDF_STRING|FDF_W,    0, NULL,             SET_Highlight },
   { "HighlightRGB", 0, FDF_POINTER|FDF_R,   0, GET_HighlightRGB, NULL },
   { "Opacity",      0, FDF_FLOAT|FDF_RW,    0, GET_Opacity,      SET_Opacity },
   { "Raised",       0, FDF_LONG|FDF_W,      0, NULL,             SET_Raised },
   { "Round",        0, FDF_LONG|FDF_W,      0, NULL,             SET_Round },
   { "Shadow",       0, FDF_STRING|FDF_W,    0, NULL,             SET_Shadow },
   { "ShadowRGB",    0, FDF_POINTER|FDF_R,   0, GET_ShadowRGB,    NULL },
   { "Sunken",       0, FDF_LONG|FDF_W,      0, NULL,             SET_Sunken },
   /*** Variable fields ***/
   { "Height",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",   0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset", 0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset", 0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*** Action Prototypes ***/

static ERROR BOX_ActionNotify(struct Box *, struct acActionNotify *);
static ERROR BOX_ClosingTag(struct Box *, APTR);
static ERROR BOX_Draw(struct Box *, APTR);
static ERROR BOX_Focus(struct Box *, APTR);
static ERROR BOX_Free(struct Box *, APTR);
static ERROR BOX_Hide(struct Box *, APTR);
static ERROR BOX_Init(struct Box *, APTR);
static ERROR BOX_LostFocus(struct Box *, APTR);
static ERROR BOX_NewObject(struct Box *, APTR);
static ERROR BOX_OpeningTag(struct Box *, APTR);
static ERROR BOX_SetUnlistedField(struct Box *, struct acSetUnlistedField *);
static ERROR BOX_Show(struct Box *, APTR);
static ERROR BOX_Timer(struct Box *, struct acTimer *);

static struct ActionArray BoxActions[] = {
   { AC_ActionNotify,     BOX_ActionNotify },
   { AC_ClosingTag,       BOX_ClosingTag },
   { AC_Draw,             BOX_Draw },
   { AC_Focus,            BOX_Focus },
   { AC_Free,             BOX_Free },
   { AC_Hide,             BOX_Hide },
   { AC_Init,             BOX_Init },
   { AC_LostFocus,        BOX_LostFocus },
   { AC_NewObject,        BOX_NewObject },
   { AC_OpeningTag,       BOX_OpeningTag },
   { AC_SetUnlistedField, BOX_SetUnlistedField },
   { AC_Show,             BOX_Show },
   { AC_Timer,            BOX_Timer },
   { NULL, NULL }
};

static ERROR ReadColour(STRING, LONG *, LONG *, LONG *);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

/*
   if (CreateObject(ID_MODULE, NULL, &modRender, NULL,
      FID_Name|TSTRING,   "render",
      FID_Version|TFLOAT, MODVERSION_RENDER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modRender, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));
*/
   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&BoxClass, NULL,
      FID_BaseClassID|TLONG, ID_BOX,
      FID_Version|TFLOAT,    VER_BOX,
      FID_Name|TSTRING,      "Box",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      BoxActions,
      FID_Fields|TPTR,       BoxFields,
      FID_Flags|TLONG,       CLF_PRIVATEONLY,
      FID_Size|TLONG,        sizeof(struct Box),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
//   if (modRender) { acFree(modRender); modRender = NULL; }
   if (BoxClass) { acFree(BoxClass); BoxClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Box: ActionNotify
*/

static ERROR BOX_ActionNotify(struct Box *Self, struct acActionNotify *Args)
{
   if (Args->ActionID IS AC_Focus) acFocus(Self);
   else if (Args->ActionID IS AC_LostFocus) acLostFocus(Self);
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: DrawBox()
*/

static void DrawBox(struct Box *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct mtDrawRectangle rect;
   struct RGB rgb, *rgbhighlight, *rgbshadow;
   ULONG colour, shadow, highlight;
   LONG XCoord, YCoord, Width, Height, XOffset, YOffset, bmpopacity;
   WORD i, j, k, round, y, border_opacity;

   if ((Self->Visible != TRUE) OR (Self->Tag)) return;

   /*** Frame testing ***/

   if ((Self->Frame) OR (Self->Frames)) {
      if (Self->Frame) {
         if (Drawable->Frame != Self->Frame) return;
      }
      else if (Self->Frames) {
         if ((Drawable->Frame > Self->TotalFrames) OR (Self->Frames[Drawable->Frame] < 1)) return;
      }
   }

   bmpopacity = Bitmap->Opacity;

   for (j=0; j < Self->TotalBoxes; j++) {
      /*** Convert relative values to their fixed equivalent ***/

      if (Self->Boxes[j].Dimensions & DMF_RELATIVE_XOFFSET) XOffset = (LONG)(Drawable->Width * Self->Boxes[j].XOffset / 100);
      else XOffset = (LONG)Self->Boxes[j].XOffset;

      if (Self->Boxes[j].Dimensions & DMF_RELATIVE_YOFFSET) YOffset = (LONG)(Drawable->Height * Self->Boxes[j].YOffset / 100);
      else YOffset = (LONG)Self->Boxes[j].YOffset;

      if (Self->Boxes[j].Dimensions & DMF_RELATIVE_X) XCoord = (LONG)(Drawable->Width * Self->Boxes[j].XCoord / 100);
      else XCoord = (LONG)Self->Boxes[j].XCoord;

      if (Self->Boxes[j].Dimensions & DMF_RELATIVE_Y) YCoord = (LONG)(Drawable->Height * Self->Boxes[j].YCoord / 100);
      else YCoord = (LONG)Self->Boxes[j].YCoord;

      /*** Calculate absolute width ***/

      if (Self->Boxes[j].Dimensions & DMF_FIXED_WIDTH) Width = (LONG)Self->Boxes[j].Width;
      else if (Self->Boxes[j].Dimensions & DMF_RELATIVE_WIDTH) Width = (LONG)(Drawable->Width * Self->Boxes[j].Width / 100);
      else if ((Self->Boxes[j].Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)) AND
               (Self->Boxes[j].Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X))) {
         Width = (LONG)(Drawable->Width - XCoord - XOffset);
      }
      else Width = NULL;

      /*** Calculate absolute height ***/

      if (Self->Boxes[j].Dimensions & DMF_FIXED_HEIGHT) Height = (LONG)Self->Boxes[j].Height;
      else if (Self->Boxes[j].Dimensions & DMF_RELATIVE_HEIGHT) Height = (LONG)(Drawable->Height * Self->Boxes[j].Height / 100);
      else if ((Self->Boxes[j].Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)) AND
               (Self->Boxes[j].Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y))) {
         Height = (LONG)(Drawable->Height - YCoord - YOffset);
      }
      else Height = NULL;

      /* If a width or height setting is null but its opposite is set,
      ** use the opposite value to create a proportional box.
      */

      if ((Width < 1) OR (Height < 1)) return;

      if (!Width) {
         if (Height) Width = Height;
         else Width = Drawable->Width;
      }

      if (!Height) {
         if (Width) Height = Width;
         else Height = Drawable->Height;
      }

      /*** X coordinate check ***/

      if ((Self->Boxes[j].Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)) AND
          (Self->Boxes[j].Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         XCoord = Drawable->Width - XOffset - Width;
      }

      /*** Y coordinate check ***/

      if ((Self->Boxes[j].Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)) AND
          (Self->Boxes[j].Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         YCoord = Drawable->Height - YOffset - Height;
      }

      Bitmap->Opacity = Self->Boxes[j].Opacity;

      if ((Self->Flags & BXF_RAISED) OR (Self->Boxes[j].Flags & BXF_RAISED)) {
         /*** Raised colours ***/
         if (Self->Boxes[j].Flags & BXF_HIGHLIGHT) rgbhighlight = &Self->Boxes[j].Highlight;
         else rgbhighlight = &Self->Highlight;
         if (Self->Boxes[j].Flags & BXF_SHADOW) rgbshadow = &Self->Boxes[j].Shadow;
         else rgbshadow = &Self->Shadow;
      }
      else if ((Self->Flags & BXF_SUNKEN) OR (Self->Boxes[j].Flags & BXF_SUNKEN)) {
         /*** Sunken colours ***/
         if (Self->Boxes[j].Flags & BXF_HIGHLIGHT) rgbshadow = &Self->Boxes[j].Highlight;
         else rgbshadow = &Self->Highlight;
         if (Self->Boxes[j].Flags & BXF_SHADOW) rgbhighlight = &Self->Boxes[j].Shadow;
         else rgbhighlight = &Self->Shadow;
      }
      else {
         /*** Single border colour ***/
         if (Self->Boxes[j].Flags & BXF_HIGHLIGHT) rgbshadow = rgbhighlight = &Self->Boxes[j].Highlight;
         else rgbshadow = rgbhighlight = &Self->Highlight;
      }

      if (Self->Brightness != 100) {
         FLOAT value;
         UWORD red, green, blue;

         value = Self->Brightness / 100.0;
         if ((red = rgbhighlight->Red * value) > 255)     red   = 255;
         if ((green = rgbhighlight->Green * value) > 255) green = 255;
         if ((blue  = rgbhighlight->Blue * value) > 255)  blue  = 255;
         highlight = PackPixel(Bitmap, red, green, blue);

         if ((red   = rgbshadow->Red * value) > 255)   red   = 255;
         if ((green = rgbshadow->Green * value) > 255) green = 255;
         if ((blue  = rgbshadow->Blue * value) > 255)  blue  = 255;
         shadow = PackPixel(Bitmap, red, green, blue);

         if (Self->Boxes[j].Flags & BXF_COLOUR) {
            if ((red   = Self->Boxes[j].Colour.Red * value) > 255)   red   = 255;
            if ((green = Self->Boxes[j].Colour.Green * value) > 255) green = 255;
            if ((blue  = Self->Boxes[j].Colour.Blue * value) > 255)  blue  = 255;
            colour = PackPixel(Bitmap, red, green, blue);
         }
         else {
            if ((red   = Self->Colour.Red * value) > 255)   red   = 255;
            if ((green = Self->Colour.Green * value) > 255) green = 255;
            if ((blue  = Self->Colour.Blue * value) > 255)  blue  = 255;
            colour = PackPixel(Bitmap, red, green, blue);
         }
      }
      else {

         highlight = PackPixel(Bitmap, rgbhighlight->Red, rgbhighlight->Green, rgbhighlight->Blue);
         shadow    = PackPixel(Bitmap, rgbshadow->Red, rgbshadow->Green, rgbshadow->Blue);

         if (Self->Boxes[j].Flags & BXF_COLOUR) colour = PackPixel(Bitmap, Self->Boxes[j].Colour.Red, Self->Boxes[j].Colour.Green, Self->Boxes[j].Colour.Blue);
         else colour = PackPixel(Bitmap, Self->Colour.Red, Self->Colour.Green, Self->Colour.Blue);
      }

      if ((Self->Boxes[j].Round >= 1) AND (Height > 3)) {
         /*** Box with rounded corners ***/

         UBYTE xmap[10][10] = {      /* This table is used to lookup the coordinates for the rounded corners */
            { 0,0,0,0,0,0,0,0,0,0 }, /* 0 */
            { 1,0,0,0,0,0,0,0,0,0 }, /* 1 */
            { 2,1,0,0,0,0,0,0,0,0 }, /* 2 */
            { 3,2,1,0,0,0,0,0,0,0 }, /* 3 */
            { 4,2,1,1,0,0,0,0,0,0 }, /* 4 */
            { 5,3,2,1,1,0,0,0,0,0 }, /* 5 */
            { 6,4,3,2,1,1,0,0,0,0 }, /* 6 */
            { 7,4,3,2,1,1,1,0,0,0 }, /* 7 */
            { 8,6,4,3,2,2,1,1,0,0 }, /* 8 */
            { 9,6,5,4,3,2,1,1,1,0 }, /* 9 */
         };

         round = Self->Boxes[j].Round;
         if (round > 9) round = 9;
         if (round > (Height-2) / 2) round = (Height-2) / 2;

            /*** Top & Bottom ***/

            for (y=0; y < round; y++) {
               if (y < Self->Thickness) {
                  /*** Draw border lines at the top and bottom of the box ***/

                  Bitmap->Opacity = Self->Boxes[j].BorderOpacity;

                  rect.XCoord = XCoord + xmap[round][y];
                  rect.YCoord = YCoord + y;
                  rect.Width  = Width - (xmap[round][y]<<1);
                  rect.Height = 1;
                  rect.Fill   = TRUE;
                  rect.Colour = highlight;
                  Action(MT_DrawRectangle, Bitmap, &rect);

                  rect.YCoord = YCoord + Height - y - 1;
                  rect.Colour = shadow;
                  Action(MT_DrawRectangle, Bitmap, &rect);
               }
               else {
                  Bitmap->Opacity = Self->Boxes[j].Opacity;

                  k = xmap[round][y];
                  if (Self->Thickness > 0) {
                     do {
                        if ((border_opacity = Self->Boxes[j].BorderOpacity) < 255) {
                           Bitmap->ReadUCRPixel(Bitmap, Bitmap->XOffset + XCoord + k, Bitmap->YOffset + YCoord + y, &rgb);
                           rgb.Red   += ((rgbhighlight->Red   - rgb.Red)   * border_opacity)>>8;
                           rgb.Green += ((rgbhighlight->Green - rgb.Green) * border_opacity)>>8;
                           rgb.Blue  += ((rgbhighlight->Blue  - rgb.Blue)  * border_opacity)>>8;
                           Bitmap->DrawUCRPixel(Bitmap, Bitmap->XOffset + XCoord + k, Bitmap->YOffset + YCoord + y, &rgb);

                           Bitmap->ReadUCRPixel(Bitmap, Bitmap->XOffset + XCoord + Width - k - 1, Bitmap->YOffset + YCoord + y, &rgb);
                           rgb.Red   += ((rgbhighlight->Red   - rgb.Red)   * border_opacity)>>8;
                           rgb.Green += ((rgbhighlight->Green - rgb.Green) * border_opacity)>>8;
                           rgb.Blue  += ((rgbhighlight->Blue  - rgb.Blue)  * border_opacity)>>8;
                           Bitmap->DrawUCRPixel(Bitmap, Bitmap->XOffset + XCoord + Width - k - 1, Bitmap->YOffset + YCoord + y, &rgb);

                           Bitmap->ReadUCRPixel(Bitmap, Bitmap->XOffset + XCoord + k, Bitmap->YOffset + YCoord + Height - y - 1, &rgb);
                           rgb.Red   += ((rgbhighlight->Red   - rgb.Red)   * border_opacity)>>8;
                           rgb.Green += ((rgbhighlight->Green - rgb.Green) * border_opacity)>>8;
                           rgb.Blue  += ((rgbhighlight->Blue  - rgb.Blue)  * border_opacity)>>8;
                           Bitmap->DrawUCRPixel(Bitmap, Bitmap->XOffset + XCoord + k, Bitmap->YOffset + YCoord + Height - y - 1, &rgb);

                           Bitmap->ReadUCRPixel(Bitmap, Bitmap->XOffset + XCoord + Width - k - 1, Bitmap->YOffset + YCoord + Height - y - 1, &rgb);
                           rgb.Red   += ((rgbshadow->Red   - rgb.Red)   * border_opacity)>>8;
                           rgb.Green += ((rgbshadow->Green - rgb.Green) * border_opacity)>>8;
                           rgb.Blue  += ((rgbshadow->Blue  - rgb.Blue)  * border_opacity)>>8;
                           Bitmap->DrawUCRPixel(Bitmap, Bitmap->XOffset + XCoord + Width - k - 1, Bitmap->YOffset + YCoord + Height - y - 1, &rgb);
                        }
                        else {
                           ActionTags(MT_DrawRGBPixel, Bitmap, XCoord + k, YCoord + y, rgbhighlight->Red, rgbhighlight->Green, rgbhighlight->Blue); /* top-left */
                           rgb.Red   = rgbhighlight->Red   + (((rgbshadow->Red   - rgbhighlight->Red)   * ((y<<8)/round))>>8);
                           rgb.Green = rgbhighlight->Green + (((rgbshadow->Green - rgbhighlight->Green) * ((y<<8)/round))>>8);
                           rgb.Blue  = rgbhighlight->Blue  + (((rgbshadow->Blue  - rgbhighlight->Blue)  * ((y<<8)/round))>>8);
                           ActionTags(MT_DrawRGBPixel, Bitmap, XCoord + Width - k - 1, YCoord + y, rgb.Red, rgb.Green, rgb.Blue); /* top-right */

                           rgb.Red   = rgbshadow->Red   + (((rgbhighlight->Red   - rgbshadow->Red)   * ((y<<8)/round))>>8);
                           rgb.Green = rgbshadow->Green + (((rgbhighlight->Green - rgbshadow->Green) * ((y<<8)/round))>>8);
                           rgb.Blue  = rgbshadow->Blue  + (((rgbhighlight->Blue  - rgbshadow->Blue)  * ((y<<8)/round))>>8);
                           ActionTags(MT_DrawRGBPixel, Bitmap, XCoord + k, YCoord + Height - y - 1, rgb.Red, rgb.Green, rgb.Blue); /* bottom-left */
                           ActionTags(MT_DrawRGBPixel, Bitmap, XCoord + Width - k - 1, YCoord + Height - y - 1, rgbshadow->Red, rgbshadow->Green, rgbshadow->Blue); /* bottom-right */
                        }
                     } while (++k < xmap[round][y-1]);
                  }

                  if ((Self->Flags & BXF_FILL) OR (Self->Boxes[j].Flags & BXF_FILL)) {
                     rect.XCoord = XCoord + k;
                     rect.YCoord = YCoord + y;
                     rect.Width  = Width - (k<<1);
                     rect.Height = 1;
                     rect.Fill   = TRUE;
                     rect.Colour = colour;
                     Action(MT_DrawRectangle, Bitmap, &rect);

                     rect.YCoord = YCoord + Height - y - 1;
                     Action(MT_DrawRectangle, Bitmap, &rect);
                  }
               }
            }

            if (Self->Thickness > 0) {
               /*** Thickness bar on either side ***/

               Bitmap->Opacity = Self->Boxes[j].BorderOpacity;

               rect.Fill   = TRUE;
               rect.XCoord = XCoord;
               rect.YCoord = YCoord + y;
               rect.Width  = Self->Thickness;
               rect.Colour = highlight;
               if ((rect.Height = Height - y - round) > 0) {
                  Action(MT_DrawRectangle, Bitmap, &rect);

                  rect.XCoord = XCoord + Width - 1;
                  rect.Colour = shadow;
                  Action(MT_DrawRectangle, Bitmap, &rect);
               }

               /*** Middle area ***/

               if ((Self->Flags & BXF_FILL) OR (Self->Boxes[j].Flags & BXF_FILL)) {
                  Bitmap->Opacity = Self->Boxes[j].Opacity;

                  rect.Fill   = TRUE;
                  rect.XCoord = XCoord + Self->Thickness;
                  rect.YCoord = YCoord + y;
                  rect.Width  = Width - (Self->Thickness<<1);
                  rect.Colour = colour;
                  if ((rect.Height = Height - y - round) > 0) {
                     Action(MT_DrawRectangle, Bitmap, &rect);
                  }
               }
            }
            else if ((Self->Flags & BXF_FILL) OR (Self->Boxes[j].Flags & BXF_FILL)) {
               /*** Middle area ***/
               Bitmap->Opacity = Self->Boxes[j].Opacity;

               rect.Fill   = TRUE;
               rect.XCoord = XCoord;
               rect.YCoord = YCoord + y;
               rect.Width  = Width;
               rect.Colour = colour;
               if ((rect.Height = Height - y - round) > 0) {
                  Action(MT_DrawRectangle, Bitmap, &rect);
               }
            }
      }
      else {
         if ((Self->Flags & BXF_FILL) OR (Self->Boxes[j].Flags & BXF_FILL)) {
            /*** Fill the Box ***/

            Bitmap->Opacity = Self->Boxes[j].Opacity;

            rect.Colour = colour;
            rect.XCoord = XCoord + Self->Thickness;
            rect.YCoord = YCoord + Self->Thickness;
            rect.Width  = Width - (Self->Thickness<<1);
            rect.Height = Height - (Self->Thickness<<1);
            rect.Fill   = TRUE;
            Action(MT_DrawRectangle, Bitmap, &rect);
         }

         /*** Draw the Box border (with highlights and shadows) ***/

         if (Self->Thickness > 0) {
            Bitmap->Opacity = Self->Boxes[j].BorderOpacity;

            for (i=0; i < Self->Thickness; i++) {

               /*** Top, Bottom ***/

               ActionTags(MT_DrawRectangle, Bitmap, XCoord+i, YCoord+i,          Width-i-i, 1, highlight, TRUE);
               ActionTags(MT_DrawRectangle, Bitmap, XCoord+i, YCoord+Height-i-1, Width-i-i, 1, shadow, TRUE);

               /*** Left, Right ***/

               ActionTags(MT_DrawRectangle, Bitmap, XCoord+i,         YCoord+i+1, 1, Height-i-i-2, highlight, TRUE);
               ActionTags(MT_DrawRectangle, Bitmap, XCoord+Width-i-1, YCoord+i+1, 1, Height-i-i-2, shadow, TRUE);
            }
         }
      }
   }

   Bitmap->Opacity = bmpopacity;
}

/*****************************************************************************
** Action: ClosingTag
*/

static ERROR BOX_ClosingTag(struct Box *Self, APTR Void)
{
   Self->Tag--;
   return(ERR_Okay);
}

/*****************************************************************************

###ACTIONS###
Name:  Draw
Short: Draws a box graphic within its drawable area.
###END###

*****************************************************************************/

static ERROR BOX_Draw(struct Box *Self, APTR Void)
{
   return(ActionMsg(AC_Draw, Self->DrawableID, NULL));
}

/*****************************************************************************

*****************************************************************************/

static ERROR BOX_Focus(struct Box *Self, APTR Void)
{
   if (Self->DarkFocus) {
      Self->Brightness = Self->DarkFocus;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR BOX_LostFocus(struct Box *Self, APTR Void)
{
   if (Self->DarkFocus) {
      Self->Brightness = 100;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Box: Free
*/

static ERROR BOX_Free(struct Box *Self, APTR Void)
{
   OBJECTPTR drawable;

   UnsubscribeTimer(Self->Head.UniqueID);

   if (Self->DrawableID) {
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, drawable, Self, NULL);
         ReleaseObject(drawable);
      }
   }

   if (Self->Pen) {
      Action(AC_Free, Self->Pen, NULL);
      Self->Pen = NULL;
   }

   if (Self->Frames) {
      FreeMemory(Self->Frames);
      Self->Frames = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Hides a box from view.
###END###

*****************************************************************************/

static ERROR BOX_Hide(struct Box *Self, APTR Void)
{
   struct acDraw draw;

   if (Self->Visible IS TRUE) {
      Self->Visible = FALSE;
      if (Self->TotalBoxes > 1) ActionMsg(AC_Draw, Self->DrawableID, NULL);
      else {
         GetFields(Self, FID_XCoord|TDOUBLE, &draw.XCoord,
                         FID_YCoord|TDOUBLE, &draw.YCoord,
                         FID_Width|TDOUBLE,  &draw.Width,
                         FID_Height|TDOUBLE, &draw.Height,
                         TAGEND);
         ActionMsg(AC_Draw, Self->DrawableID, &draw);
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Box: Init
*/

static ERROR BOX_Init(struct Box *Self, APTR Void)
{
   OBJECTID OwnerID;
   OBJECTPTR drawable;
   struct Variable Value;

   if (!Self->DrawableID) {
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (OwnerID) Self->DrawableID = OwnerID;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   /* If dimension settings are missing (e.g. if it is impossible to
   ** determine width, height or a coordinate), then we set the missing
   ** fields to maximum possible values.
   */

   Value.Type   = FD_DOUBLE;
   Value.Double = 0;
   if (!(Self->Boxes[0].Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_X(Self, &Value);
   }

   if (!(Self->Boxes[0].Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_Y(Self, &Value);
   }

   if (!(Self->Boxes[0].Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_XOffset(Self, &Value);
   }

   if (!(Self->Boxes[0].Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_YOffset(Self, &Value);
   }

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, drawable, Self, &DrawBox);

      if (Self->DarkFocus) {
         SubscribeActionTags(drawable, Self, AC_Focus, AC_LostFocus, TAGEND);
      }
      ReleaseObject(drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_ExclusiveDenied));

   /*** If an effect has been specified, initialise the timer ***/

   if (Self->Effect) {
      if (Self->Effect IS BFX_FADE) Self->Boxes[0].Opacity = 0;
      else if (Self->Effect IS BFX_FADEOUT) Self->Boxes[0].Opacity = 255;

      if ((!Self->Delay) AND (!Self->Frame) AND (!Self->Frames)) {
         /*** Start the effect now if their are no predetermined triggers ***/
         Self->Delay = -1;
         Self->EffectActive = 1;
      }

      SubscribeTimer(Self->Head.UniqueID, 20);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Box: NewObject
*/

static ERROR BOX_NewObject(struct Box *Self, APTR Void)
{
   WORD i;

   Self->TotalBoxes = 1;
   Self->Visible    = TRUE;
   Self->Duration   = 1000;
   Self->Brightness = 100;

   for (i=0; i < MAX_BOXES; i++) {
      Self->Boxes[i].Opacity = 255;
      Self->Boxes[i].BorderOpacity = 255;
   }

   /*** Shadow colour ***/
   Self->Shadow.Red   = 64;
   Self->Shadow.Green = 64;
   Self->Shadow.Blue  = 64;

   /*** Internal colour ***/
   Self->Colour.Red   = 128;
   Self->Colour.Green = 128;
   Self->Colour.Blue  = 128;

   /*** Highlight colour ***/
   Self->Highlight.Red   = 192;
   Self->Highlight.Green = 192;
   Self->Highlight.Blue  = 192;

   return(ERR_Okay);
}

/*****************************************************************************
** Action: OpeningTag
*/

static ERROR BOX_OpeningTag(struct Box *Self, APTR Void)
{
   Self->Tag++;
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Shows a box.
###END###

*****************************************************************************/

static ERROR BOX_Show(struct Box *Self, APTR Void)
{
   struct acDraw draw;

   if (Self->Visible IS FALSE) {
      Self->Visible = TRUE;

      if (Self->TotalBoxes > 1) ActionMsg(AC_Draw, Self->DrawableID, NULL);
      else {
         GetFields(Self, FID_XCoord|TDOUBLE, &draw.XCoord,
                         FID_YCoord|TDOUBLE, &draw.YCoord,
                         FID_Width|TDOUBLE,  &draw.Width,
                         FID_Height|TDOUBLE, &draw.Height,
                         TAGEND);
         ActionMsg(AC_Draw, Self->DrawableID, &draw);
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  SetUnlistedField
Short: Used to set special box fields.
###END###

*****************************************************************************/

static ERROR BOX_SetUnlistedField(struct Box *Self, struct acSetUnlistedField *Args)
{
   LONG i;

   if (StrCompare("highlight", Args->Field, 9, NULL) IS ERR_Okay) {
      i = StrToInt(Args->Field+9);
      ReadColour(Args->Value, &Self->Boxes[i].Highlight.Red, &Self->Boxes[i].Highlight.Green, &Self->Boxes[i].Highlight.Blue);
      Self->Boxes[i].Flags |= BXF_HIGHLIGHT;
      if (!Self->Thickness) Self->Thickness = 1;
   }
   else if (StrCompare("shadow", Args->Field, 6, NULL) IS ERR_Okay) {
      i = StrToInt(Args->Field+6);
      ReadColour(Args->Value, &Self->Boxes[i].Shadow.Red, &Self->Boxes[i].Shadow.Green, &Self->Boxes[i].Shadow.Blue);
      Self->Boxes[i].Flags |= BXF_SHADOW;
      if (!Self->Thickness) Self->Thickness = 1;
   }
   else if (StrCompare("col", Args->Field, 3, NULL) IS ERR_Okay) {
      for (i=0; (Args->Field[i]) AND ((Args->Field[i] < '0') OR (Args->Field[i] > '9')); i++);
      if (Args->Field[0]) {
         i = StrToInt(Args->Field+i);
         ReadColour(Args->Value, &Self->Boxes[i].Colour.Red, &Self->Boxes[i].Colour.Green, &Self->Boxes[i].Colour.Blue);
         Self->Boxes[i].Flags |= BXF_COLOUR|BXF_FILL;
      }
   }
   else if (StrCompare("border", Args->Field, 6, NULL) IS ERR_Okay) {
      i = StrToInt(Args->Field+6);
      ReadColour(Args->Value, &Self->Boxes[i].Highlight.Red, &Self->Boxes[i].Highlight.Green, &Self->Boxes[i].Highlight.Blue);
      ReadColour(Args->Value, &Self->Boxes[i].Shadow.Red, &Self->Boxes[i].Shadow.Green, &Self->Boxes[i].Shadow.Blue);
      Self->Boxes[i].Flags |= BXF_HIGHLIGHT|BXF_SHADOW;
      if (!Self->Thickness) Self->Thickness = 1;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Box: Timer
*/

static ERROR BOX_Timer(struct Box *Self, struct acTimer *Args)
{
   struct acDraw draw;
   OBJECTPTR drawable;
   LONG frame, clear;

   clear = FALSE;
   if (Self->Delay > 0) {
      if (Args) Self->Delay -= Args->TimeElapsed;
      if (Self->Delay > 0) return(ERR_Okay);

      Self->EffectActive = 1;
      Self->Delay = -1;
      SubscribeTimer(Self->Head.UniqueID, 20); /* Resubscribe to reset the total elapsed time */
   }
   else if ((Self->Frame) OR (Self->Frames)) {
      if ((!Self->EffectActive) OR (Self->EffectActive IS 2)) {
         if (AccessObject(Self->DrawableID, 2, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Frame, FT_LONG, &frame);
            ReleaseObject(drawable);

            /*** Reset the effect if the animation has looped back to frame zero ***/

            if (Self->EffectActive IS 2) {
               if (frame < Self->CurrentFrame) {
                  Self->EffectActive = 0;
                  SubscribeTimer(Self->Head.UniqueID, 20); /* Resubscribe to reset the total elapsed time */
               }
            }

            if (!Self->EffectActive) {
               if (Self->Frame) {
                  if (frame IS Self->Frame) {
                     Self->EffectActive = 1;
                     clear = TRUE;
                     SubscribeTimer(Self->Head.UniqueID, 20);
                  }
               }
               else if ((frame <= Self->TotalFrames) AND (Self->Frames[frame] >= 1)) {
                   Self->EffectActive = 1;
                   clear = TRUE;
                   SubscribeTimer(Self->Head.UniqueID, 20);
                }
            }

            Self->CurrentFrame = frame;
         }
         else return(ERR_AccessObject);
      }

      if ((!Self->EffectActive) OR (Self->EffectActive IS 2)) return(ERR_Okay);
   }

   /*** Effect management routines ***/

   if ((Self->Effect IS BFX_FADE) OR (Self->Effect IS BFX_FADEOUT)) {
      if (Self->Effect IS BFX_FADE) {
         /*** Fade in ***/
         Self->Boxes[0].Opacity = (Args->TotalElapsed<<8) / Self->Duration;
         if (Self->Boxes[0].Opacity >= 255) {
            Self->Boxes[0].Opacity = 255;
            if (!(Self->Flags & BXF_REPEATEFFECT)) UnsubscribeTimer(Self->Head.UniqueID);
         }
      }
      else {
         /*** Fade out ***/
         Self->Boxes[0].Opacity = 255 - ((Args->TotalElapsed<<8) / Self->Duration);
         if (Self->Boxes[0].Opacity < 0) {
            Self->Boxes[0].Opacity = 0;
            if (!(Self->Flags & BXF_REPEATEFFECT)) UnsubscribeTimer(Self->Head.UniqueID);
         }
      }

      if (!(Self->Flags & BXF_NOFXDRAW)) {
         GetFields(Self, FID_XCoord|TDOUBLE, &draw.XCoord,
                         FID_YCoord|TDOUBLE, &draw.YCoord,
                         FID_Width|TDOUBLE,  &draw.Width,
                         FID_Height|TDOUBLE, &draw.Height,
                         TAGEND);
         ActionMsg(AC_Draw, Self->DrawableID, &draw);
      }
   }
   else {
      /*** Unknown effect type ***/
      Self->Effect = NULL;
      UnsubscribeTimer(Self->Head.UniqueID);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ReadByte()
** Short:    Reads any double digit hex value and converts it into a number.
*/

static WORD ReadByte(STRING Str)
{
   WORD Value;
   if (!*Str) return(0);
   if ((*Str >= '0') AND (*Str <= '9')) Value = (*Str - '0') * 16;
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value = ((*Str - 'A')+10) * 16;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value = ((*Str - 'a')+10) * 16;
   else Value = 0;
   Str++;
   if ((*Str >= '0') AND (*Str <= '9')) Value += (*Str - '0');
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value += (*Str - 'A') + 10;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value += (*Str - 'a') + 10;
   return(Value);
}

static ERROR ReadColour(STRING Colour, LONG *Red, LONG *Green, LONG *Blue)
{
   if (*Colour IS '#') {
      Colour++;
      *Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Blue  = ReadByte(Colour);
   }
   else {
      *Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Blue = StrToInt(Colour);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Border
Short:  String-based field for setting the border colour of a Box object.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>To give a Box a plain-coloured border on all four sides, set this field to
the text-based colour value that you want to use.  If the Highlight or Shadow
fields are set, giving the Box a border will over-ride those values with the
new border colour.</p>

<p>To set the thickness of the border, write a value to the Thickness field.</p>

###SEE ALSO###
Field: Thickness

###END###

*****************************************************************************/

static ERROR SET_Border(struct Box *Self, STRING Value)
{
   SET_Highlight(Self, Value);
   SET_Shadow(Self, Value);
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   BorderOpacity
Short:  Determines the level of translucency applied to box borders.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>This field determines the translucency level of a box graphic.  The
default setting is 100%, which means that the box will be solid.  Any
other value that you set here will alter the impact of a box graphic over
the destination drawable.  High values will retain the boldness of the box,
while low values can render it close to invisible.</p>

<p>Please note that the use of translucency will always have an impact on the
time it normally takes to draw a box.</p>

###END###

****************************************************************************/

static ERROR GET_BorderOpacity(struct Box *Self, FLOAT *Value)
{
   *Value = Self->Boxes[0].BorderOpacity * 100 / 255;
   return(ERR_Okay);
}

static ERROR SET_BorderOpacity(struct Box *Self, FLOAT *Value)
{
   WORD i;

   if (*Value < 0) Self->Boxes[0].BorderOpacity = 0;
   else if (*Value > 100) Self->Boxes[0].BorderOpacity = 255;
   else Self->Boxes[0].BorderOpacity = *Value * 255 / 100;

   for (i=1; i < MAX_BOXES; i++) Self->Boxes[i].BorderOpacity = Self->Boxes[0].BorderOpacity;

   if (Self->Boxes[0].BorderOpacity >= 255) Self->Flags &= ~BXF_BORDEROPACITY;
   else Self->Flags |= BXF_BORDEROPACITY;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Boxes
Short:  Special field to create multiple boxes using one object.
Type:   STRING (Unlisted)
Status: Write

###DESCRIPTION###
<p>This special field allows you to create multiple box definitions using
a single Box object.  This can help you save memory in cases where you might
otherwise need to create many Box objects to get the effect that you want
to achieve.  Setting this field requires that you follow some basic rules
and use the correct symbols, or you may get unpredictable results.</p>

<p>Basic use requires that you specify the coordinates and dimensions for
each box in sets of four numbers - that is, the X and Y coordinate followed
by either the Width and Height or XOffset and YOffset.  To aid you in your
visualisation of these numbers, you may insert brackets and commas to
organise each box definition into a group.  The following example illustrates
the basic creation of two box graphics:</p>

<pre>
  "(0,0,20,20) (5,5,10,10)"
</pre>

<p>This would result in a box that is 20x20 pixels wide at position (0,0),
and a box that is 10x10 pixels wide at position (5,5).  If you need to
set an offset or percentage based value, then you will need to use the
exclamation and/or percentage symbols.  This next example will create
a box that surrounds its contained area with two box definitions:</p>

<pre>
  "(0,0,!0,!0) (1,1,!1,!1)"
</pre>

<p>This example uses percentages to create two boxes:</p>

<pre>
  "(0%,0%,20%,20%) (5%,5%,10%,10%)"
</pre>

<p>The only issues left are in the setting of colours, and whether
or not individual boxes should be given a raised or sunken appearance.
Creating a raised or sunken box requires you to set an "R" or "S"
character prior to a box definition.  The following example creates two
separate raised and sunken boxes:</p>

<pre>
  "r(0,0,20,20) s(5,5,10,10)"
</pre>

<p>In order to set colours for individual boxes, the unlisted field
mechanism needs to be used, and you need to know the correct index for
each box.  To demonstrate, setting the border for the second box in
the list would require us to set the field "Border1" (indexing starts
from zero, so the second box is #1).  Other colour indexes are supported,
including the Highlight, Shadow and Colour fields.  If a colour is not
set for a particular box definition, the box will remain unfilled.
The final example demonstrates the creation of four boxes with their
own individual border colours:</p>

<pre>
  &lt;box boxes="(0,0,!0,!0) (1,1,!1,!1) (2,2,!2,!2) (3,3,!3,!3)"
    border0="#5e4fcd" border1="#8793e8" border2="#6F7fdd" border3="#5e4fcd"/&gt;
</pre>

###END###

*****************************************************************************/

static ERROR SET_Boxes(struct Box *Self, STRING Boxes)
{
   STRING value;
   WORD offset, i, relative;

   if (!Boxes) return(ERR_Failed);

   i = 0;
   while (*Boxes) {
      /*** X Coordinate / X Offset ***/

      offset   = FALSE;
      relative = FALSE;
      while ((*Boxes) AND ((*Boxes < '0') OR (*Boxes > '9'))) {
         if (*Boxes IS '!') offset = TRUE;
         if ((*Boxes IS 'f') OR (*Boxes IS 'F')) Self->Boxes[i].Flags |= BXF_FILL;
         if ((*Boxes IS 'r') OR (*Boxes IS 'R')) Self->Boxes[i].Flags |= BXF_RAISED;
         if ((*Boxes IS 's') OR (*Boxes IS 'S')) Self->Boxes[i].Flags |= BXF_SUNKEN;
         Boxes++;
      }
      if (!*Boxes) break;

      value = Boxes;
      while ((*Boxes >= '0') AND (*Boxes <= '9')) Boxes++;
      if (*Boxes IS '%') relative = TRUE;

      if (offset IS TRUE) {
         if (relative IS TRUE) Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
         else Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
         Self->Boxes[i].XOffset = (FLOAT)StrToInt(value);
      }
      else {
         if (relative IS TRUE) Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
         else Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
         Self->Boxes[i].XCoord = (FLOAT)StrToInt(value);
      }

      if (!*Boxes) break;

      /*** Y Coordinate / Y Offset ***/

      offset = FALSE;
      relative = FALSE;
      while ((*Boxes) AND ((*Boxes < '0') OR (*Boxes > '9'))) {
         if (*Boxes IS '!') offset = TRUE;
         Boxes++;
      }
      if (!*Boxes) break;

      value = Boxes;
      while ((*Boxes >= '0') AND (*Boxes <= '9')) Boxes++;
      if (*Boxes IS '%') relative = TRUE;

      if (offset IS TRUE) {
         if (relative IS TRUE) Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
         else Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
         Self->Boxes[i].YOffset = (FLOAT)StrToFloat(value);
      }
      else {
         if (relative IS TRUE) Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
         else Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
         Self->Boxes[i].YCoord = (FLOAT)StrToFloat(value);
      }

      if (*Boxes IS NULL) break;

      /*** Width / X Offset ***/

      offset   = FALSE;
      relative = FALSE;
      while ((*Boxes) AND ((*Boxes < '0') OR (*Boxes > '9'))) {
         if (*Boxes IS '!') offset = TRUE;
         Boxes++;
      }
      if (!*Boxes) break;

      value = Boxes;
      while ((*Boxes >= '0') AND (*Boxes <= '9')) Boxes++;
      if (*Boxes IS '%') relative = TRUE;

      if (offset IS TRUE) {
         if (relative IS TRUE) Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
         else Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
         Self->Boxes[i].XOffset = (FLOAT)StrToFloat(value);
      }
      else {
         if (relative IS TRUE) Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
         else Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
         Self->Boxes[i].Width = (FLOAT)StrToFloat(value);
      }

      if (*Boxes IS NULL) break;

      /*** Height / Y Offset ***/

      offset   = FALSE;
      relative = FALSE;
      while ((*Boxes) AND ((*Boxes < '0') OR (*Boxes > '9'))) {
         if (*Boxes IS '!') offset = TRUE;
         Boxes++;
      }
      if (!*Boxes) break;

      value = Boxes;
      while ((*Boxes >= '0') AND (*Boxes <= '9')) Boxes++;
      if (*Boxes IS '%') relative = TRUE;

      if (offset IS TRUE) {
         if (relative IS TRUE) Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
         else Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
         Self->Boxes[i].YOffset = (FLOAT)StrToFloat(value);
      }
      else {
         if (relative IS TRUE) Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
         else Self->Boxes[i].Dimensions = (Self->Boxes[i].Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
         Self->Boxes[i].Height = (FLOAT)StrToFloat(value);
      }

      i++;
   }

   Self->TotalBoxes = i;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Brightness
Short:  Alters the brightness level of the box colours.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>The brightness of the colours in the box graphic can be dynamically
altered through this field.  By default the brightness value is set to 100%.
Lowering the brightness will darken the colours, while increasing the level
will brighten them.</p>

<p>If the ACTIVEDRAW flag is in use, the box will automatically redraw
itself when the Brightness field is set.</p>

###END###

*****************************************************************************/

static ERROR SET_Brightness(struct Box *Self, FLOAT *Value)
{
   Self->Brightness = *Value;
   if (Self->Flags & BXF_ACTIVEDRAW) DelayMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ColourRGB
Short:  Defines the internal colour of a box in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The colour drawn inside of a box can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ColourRGB(struct Box *Self, struct RGB *Value)
{
   Value->Red   = Self->Colour.Red;
   Value->Green = Self->Colour.Green;
   Value->Blue  = Self->Colour.Blue;
   Value->Alpha = Self->Colour.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Colour
Short:  String-based field for setting the colour inside of the box.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The internal colour of a box can be declared by writing to this field.
The colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of "#ff0000" or "255,0,0" would be
valid.</p>

<p>If you create a Box without setting the Colour field, the result will be
an unfilled rectangle.  The colour is also intrinsically linked to the Fill
field - if you want to switch the colour off temporarily, you can set the
Fill field to FALSE.</p>

###SEE ALSO###
Fields: Border, Highlight, Shadow, Fill

###END###

*****************************************************************************/

static ERROR SET_Colour(struct Box *Self, STRING Colour)
{
   LONG fill;
   if (Colour) {
      ReadColour(Colour, &Self->Colour.Red, &Self->Colour.Green, &Self->Colour.Blue);
      fill = TRUE;
      SET_Fill(Self, &fill);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  Defines the drawable area for the Box graphics.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In most cases, when you create a new box object it should be contained by
a  drawable area, such as a Render object.  If you do this correctly then
the box will detect the drawable area and set the Drawable field to the
correct object.  However, there may be occasions where you want to
initialise a box to a non-graphical container.  If this is the case then you
must set the Drawable field to point to a valid drawable area, such as a
Render object.</p>

###END###

*****************************************************************************/

/****************************************************************************

###FIELD###
Name:   Fill
Short:  Set to TRUE or FALSE to alter the filling of a box.
Type:   BOOLEAN
Status: Write

###DESCRIPTION###
<p>Manipulation of the Fill field can provide an easy mechanism for turning
the internal colour of the box on and off.  By default a box object will
manage filling based on your colour specifications, but you can force the
issue by writing directly to this field.</p>

###END###

****************************************************************************/

static ERROR GET_Fill(struct Box *Self, LONG *Value)
{
   if (Self->Flags & BXF_FILL) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Fill(struct Box *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & BXF_FILL) return(ERR_Okay);
      else Self->Flags |= BXF_FILL;
   }
   else Self->Flags &= ~BXF_FILL;

   if (Self->Flags & BXF_ACTIVEDRAW) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_Flags(struct Box *Self, LONG *Value)
{
   /*** Don't touch the upper word as it contains our private flags ***/
   Self->Flags = (Self->Flags & 0xffff0000) | ((*Value) & 0xffff);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Limits the drawing of a box to a single frame.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>Setting the Frame field to any value other than zero will force a box to
be drawn only when the drawable's frame matches the specified value.  For
instance, if the drawable container has a frame setting of 2, and the box has
a frame of 1, then no graphics will be drawn to represent the box as the
numbers do not match.</p>

###SEE ALSO###
Field: Frames

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frames
Short:  Determines a sequence of frames that a box graphic may be drawn to.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>For long animation sequences you can set the Frames field to determine
individual frames that a box graphic should be drawn to.  For instance,
if a drawable has been assigned 20 animation frames and you want a box
graphic to appear in frames 5, 6, 7, 8, 17 and 19, you would use this
frame specification:</p>

<pre>
   5-8,17,19
</pre>

<p>Individual frame sets are separated by any non-numeric character such as
a comma, and ranges can be specified through the '-' character.</p>

<p>Setting the Frames field will override any setting that you may have
previously set in the <a href="#Frame">Frame</a> field.</p>

###SEE ALSO###
Field: Frame

###END###

*****************************************************************************/

static ERROR SET_Frames(struct Box *Self, STRING Value)
{
   STRING str;
   LONG frame, end;

   if ((!Value) OR (!(*Value))) return(ERR_Args);

   /*** Calculate the size of the frame array that we are going to need ***/

   Self->TotalFrames = 0;
   str = Value;
   while (*str) {
      if ((*str >= '0') AND (*str <= '9')) {
         frame = StrToInt(str);
         while ((*str >= '0') AND (*str <= '9')) str++;
         if (frame > Self->TotalFrames) Self->TotalFrames = frame;
      }
      else str++;
   }

   Self->TotalFrames++; /* Add an extra frame because we start from #1, not zero */

   /*** Allocate the frame array and record the sequence ***/

   if (AllocMemory((Self->TotalFrames+1) * sizeof(WORD), MEM_DATA, &Self->Frames, NULL) IS ERR_Okay) {
      str = Value;
      while (*str) {
         if ((*str >= '0') AND (*str <= '9')) {
            if ((frame = StrToInt(str)) > 0) {
               while ((*str >= '0') AND (*str <= '9')) str++;
               while ((*str) AND (*str <= 0x20)) str++;
               if (*str IS '-') {
                  while ((*str) AND ((*str < '0') OR (*str > '9'))) str++;
                  if ((end = StrToInt(str)) > 0) {
                     while (frame != end) {
                        Self->Frames[frame] = 1;
                        if (frame < end) frame++;
                        else frame--;
                     }
                     Self->Frames[frame] = 1;
                  }
                  while ((*str >= '0') AND (*str <= '9')) str++;
               }
               else Self->Frames[frame] = 1;
            }
            else while ((*str >= '0') AND (*str <= '9')) str++;
         }
         else str++;
      }

      Self->Frames[Self->TotalFrames] = -1; /* Sequence termination */
   }
   else return(ERR_AllocMemory);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  The height of a box graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A Box can be given a fixed height by setting this field to the desired
value.  A percentage may be specified if the FD_PERCENT flag is used when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct Box *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerheight, ycoord, value;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ObjectError(Self, ERH_GetField, ERR_GetField));

      if (Self->Boxes[0].Dimensions & DMF_FIXED_HEIGHT) value = Self->Boxes[0].Height;
      else if (Self->Boxes[0].Dimensions & DMF_RELATIVE_HEIGHT) {
         value = Self->Boxes[0].Height * containerheight / 100;
      }
      else if ((Self->Boxes[0].Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Boxes[0].Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Boxes[0].Dimensions & DMF_FIXED_Y) ycoord = Self->Boxes[0].YCoord;
         else ycoord = containerheight * Self->Boxes[0].YCoord / 100;
         if (Self->Boxes[0].Dimensions & DMF_FIXED_YOFFSET) value = containerheight - ycoord - Self->Boxes[0].YOffset;
         else value = containerheight - ycoord - (containerheight * Self->Boxes[0].YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct Box *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Box] A height of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Boxes[0].Height = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Box] A height of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Boxes[0].Height = Value->Large;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
   else Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HighlightRGB
Short:  Defines the highlight colour of the box in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The highlight colour of a Box can be read in RGB format from this
field.</p>

###SEE ALSO###
Field: Highlight, ShadowRGB, ColourRGB

###END###

*****************************************************************************/

static ERROR GET_HighlightRGB(struct Box *Box, struct RGB *Colour)
{
   Colour->Red   = Box->Highlight.Red;
   Colour->Green = Box->Highlight.Green;
   Colour->Blue  = Box->Highlight.Blue;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Highlight
Short:  String-based field for setting the highlight colour of a box object.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>If a Box object should be highlighted around the edges, set this field
to the desired colour for the highlighting.  The colour specification must be
in the standard hexadecimal or decimal separated format.</p>

<p>Passing a NULL value to this field will turn off highlighting.</p>

###SEE ALSO###
Field: HighlightRGB, Shadow

###END###

*****************************************************************************/

static ERROR SET_Highlight(struct Box *Self, STRING Colour)
{
   if (!Colour) Self->Thickness = 0;
   else {
      ReadColour(Colour, &Self->Highlight.Red, &Self->Highlight.Green, &Self->Highlight.Blue);
      if (!Self->Thickness) Self->Thickness = 1;
   }
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Opacity
Short:  Determines the level of translucency applied to a box.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>This field determines the translucency level of a box graphic.  The
default setting is 100%, which means that the box will be solid.  Any
other value that you set here will alter the impact of a box graphic over
the destination drawable.  High values will retain the boldness of the box,
while low values can render it close to invisible.</p>

<p>Please note that the use of translucency will always have an impact on the
time it normally takes to draw a box.</p>

###SEE ALSO###
Field: BorderOpacity

###END###

****************************************************************************/

static ERROR GET_Opacity(struct Box *Self, FLOAT *Value)
{
   *Value = Self->Boxes[0].Opacity * 100 / 255;
   return(ERR_Okay);
}

static ERROR SET_Opacity(struct Box *Self, FLOAT *Value)
{
   WORD i;

   if (*Value < 0) Self->Boxes[0].Opacity = 0;
   else if (*Value > 100) Self->Boxes[0].Opacity = 255;
   else Self->Boxes[0].Opacity = *Value * 255 / 100;
   Self->Boxes[0].BorderOpacity = Self->Boxes[0].Opacity;

   for (i=1; i < MAX_BOXES; i++) {
      Self->Boxes[i].Opacity = Self->Boxes[0].Opacity;
      Self->Boxes[i].BorderOpacity = Self->Boxes[0].BorderOpacity;
   }

   if (Self->Boxes[0].Opacity >= 255) Self->Flags &= ~BXF_OPACITY;
   else Self->Flags |= BXF_OPACITY;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Raised
Short:  If set to TRUE the box will appear to be raised into the foreground.
Type:   BOOLEAN
Status: Write

###DESCRIPTION###
<p>If you have set the Highlight and Shadow fields of a box object then you
will need to decide whether or not the box should be given a sunken or raised
effect when it is drawn.  To give it a raised effect you will need to set this
field to TRUE, if not then you should set the Sunken field.</p>

###SEE ALSO###
Field: Sunken

###END###

*****************************************************************************/

static ERROR SET_Raised(struct Box *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & BXF_RAISED) return(ERR_Okay);
      Self->Flags = (Self->Flags & ~BXF_SUNKEN) | BXF_RAISED;
   }
   else Self->Flags &= ~BXF_RAISED;

   if (Self->Flags & BXF_ACTIVEDRAW) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:  Round
Short: Pixel specification for rounding at each corner of the box.
Type:  LONG
Status: Set

###DESCRIPTION###
<p>Rounded corners can be enabled by setting this field to a value of at least
two.</p>

###END###

*****************************************************************************/

static ERROR SET_Round(struct Box *Self, LONG *Value)
{
   Self->Boxes[0].Round = *Value;
   if (Self->Boxes[0].Round < 2) Self->Boxes[0].Round = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Shadow
Short:  String-based field for setting the shadow colour of a Box object.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>If a Box object should be shadowed around the edges, set this field
to the desired colour for the shadowing.  The colour specification must be
in the standard hexadecimal or decimal separated format.</p>

<p>Passing a NULL value to this field will turn off shadowing.</p>

###SEE ALSO###
Field: ShadowRGB, Highlight

###END###

*****************************************************************************/

static ERROR SET_Shadow(struct Box *Self, STRING Colour)
{
   if (!Colour) Self->Thickness = 0;
   else {
      ReadColour(Colour, &Self->Shadow.Red, &Self->Shadow.Green, &Self->Shadow.Blue);
      if (!Self->Thickness) Self->Thickness = 1;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ShadowRGB
Short:  Defines the shadow colour of a box in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The shadow colour of a Box can be read in RGB format from this field.</p>

###SEE ALSO###
Fields: Shadow, HighlightRGB

###END###

*****************************************************************************/

static ERROR GET_ShadowRGB(struct Box *Self, struct RGB *Value)
{
   Value->Red   = Self->Shadow.Red;
   Value->Green = Self->Shadow.Green;
   Value->Blue  = Self->Shadow.Blue;
   Value->Alpha = Self->Shadow.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Sunken
Short:  Set to TRUE to make a box appear to sink into the background.
Type:   BOOLEAN
Status: Write

###DESCRIPTION###
<p>If you have set the Highlight and Shadow fields of a Box object then you
will need to decide whether or not the Box should be given a sunken or raised
effect when it is drawn.  To give it a sunken effect you will need to set this
field to TRUE, if not then you should set the Raised field.</p>

###SEE ALSO###
Field: Raised

###END###

*****************************************************************************/

static ERROR SET_Sunken(struct Box *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & BXF_SUNKEN) return(ERR_Okay);
      Self->Flags = (Self->Flags & ~BXF_RAISED) | BXF_SUNKEN;
   }
   else Self->Flags &= ~BXF_SUNKEN;

   if (Self->Flags & BXF_ACTIVEDRAW) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Thickness
Short:  Defines the thickness of a box's border.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The pixel-based thickness of the Box' border is defined by this field.  A
minimum value of 0 will prevent a border from being drawn, effectively
switching off the functionality of the Highlight, Shadow and Border fields.
The default border thickness is a value of 1.</p>

###SEE ALSO###
Field: Border, Highlight, Shadow

###END###

*****************************************************************************/

static ERROR SET_Thickness(struct Box *Self, LONG *Thickness)
{
   if (Self->Thickness != *Thickness) {
      if ((*Thickness < 0) OR (*Thickness > 20)) return(ERR_OutOfRange);
      Self->Thickness = *Thickness;

      if (Self->Flags & BXF_ACTIVEDRAW) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TotalBoxes
Short:  Defines the total amount of boxes to be drawn by a box object.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>The TotalBoxes field can be dynamically adjusted if you wish to lower
or increase the amount of boxes being drawn.  This feature works
in situations where a box object contains multiple box definitions.</p>

<p>You cannot set a value less than one or greater than the amount of
boxes that you have already defined.</p>

###END###

*****************************************************************************/

static ERROR SET_TotalBoxes(struct Box *Self, LONG *Value)
{
   LONG i;

   if (*Value <= 0) return(ERR_OutOfRange);
   if (*Value > MAX_BOXES) return(ERR_OutOfRange);

   if (!Self->Boxes[*Value].Dimensions) {
      for (i=0; Self->Boxes[i].Dimensions; i++);
      Self->TotalBoxes = i;
   }
   else Self->TotalBoxes = *Value;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  The width of a box's graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>If the Box is to be drawn with a fixed width, set this field to the desired
value.  A percentage may be specified if the FD_PERCENT flag is used when
setting the field.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

static ERROR GET_Width(struct Box *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, xcoord, value;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ObjectError(Self, ERH_GetField, ERR_GetField));

      if (Self->Boxes[0].Dimensions & DMF_FIXED_WIDTH) value = Self->Boxes[0].Width;
      else if (Self->Boxes[0].Dimensions & DMF_RELATIVE_WIDTH) {
         value = Self->Boxes[0].Width * containerwidth / 100;
      }
      else if ((Self->Boxes[0].Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Boxes[0].Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Boxes[0].Dimensions & DMF_FIXED_X) xcoord = Self->Boxes[0].XCoord;
         else xcoord = containerwidth * Self->Boxes[0].XCoord / 100;
         if (Self->Boxes[0].Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - xcoord - Self->Boxes[0].XOffset;
         else value = containerwidth - xcoord - (containerwidth * Self->Boxes[0].XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Width(struct Box *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Box] A width of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Boxes[0].Width = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Box] A width of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Boxes[0].Width = Value->Large;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
   else Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of a box.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of a Box can be fixed to an absolute coordinate by
writing a value to the XCoord field.  If you use the FD_PERCENT flag then
you may write a relative/percentage based value to the field.  Negative
values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Box *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Boxes[0].Dimensions & DMF_FIXED_X) value = Self->Boxes[0].XCoord;
      else if (Self->Boxes[0].Dimensions & DMF_RELATIVE_X) {
         value = Self->Boxes[0].XCoord * containerwidth / 100;
      }
      else if ((Self->Boxes[0].Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Boxes[0].Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Boxes[0].Dimensions & DMF_FIXED_WIDTH) width = Self->Boxes[0].Width;
         else width = containerwidth * Self->Boxes[0].Width / 100;
         if (Self->Boxes[0].Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - width - Self->Boxes[0].XOffset;
         else value = containerwidth - width - (containerwidth * Self->Boxes[0].XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Box *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->Boxes[0].XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->Boxes[0].XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
   else Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horizontal offset of a box.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the Box will be drawn
from that X coordinate up to the width of the container, minus the value
given in the XOffset.  This means that the width of the Box is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the Box will be drawn at an X coordinate calculated from the formula
"XCoord = ContainerWidth - BoxWidth - XOffset".</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Box *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Boxes[0].Dimensions & DMF_FIXED_XOFFSET) value = Self->Boxes[0].XOffset;
      else if (Self->Boxes[0].Dimensions & DMF_RELATIVE_XOFFSET) {
         value = Self->Boxes[0].XOffset * containerwidth / 100;
      }
      else if ((Self->Boxes[0].Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Boxes[0].Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Boxes[0].Dimensions & DMF_FIXED_WIDTH) width = Self->Boxes[0].Width;
         else width = containerwidth * Self->Boxes[0].Width / 100;
         if (Self->Boxes[0].Dimensions & DMF_FIXED_X) value = containerwidth - (Self->Boxes[0].XCoord + width);
         else value = containerwidth - ((Self->Boxes[0].XCoord * containerwidth / 100) + width);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Box *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->Boxes[0].XOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->Boxes[0].XOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
   else Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical coordinate of a box.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of a Box can be fixed to an absolute coordinate by
writing a value to the YCoord field.  If you use the FD_PERCENT flag then
you may write a relative/percentage based value to the field.  Negative
values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct Box *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Boxes[0].Dimensions & DMF_FIXED_Y) value = Self->Boxes[0].YCoord;
      else if (Self->Boxes[0].Dimensions & DMF_RELATIVE_Y) {
         value = Self->Boxes[0].YCoord * containerheight / 100;
      }
      else if ((Self->Boxes[0].Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Boxes[0].Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Boxes[0].Dimensions & DMF_FIXED_HEIGHT) height = Self->Boxes[0].Height;
         else height = containerheight * Self->Boxes[0].Height / 100;
         if (Self->Boxes[0].Dimensions & DMF_FIXED_YOFFSET) value = containerheight - height - Self->Boxes[0].YOffset;
         else value = containerheight - height - (containerheight * Self->Boxes[0].YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Box *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->Boxes[0].YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->Boxes[0].YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
   else Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset of a box.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the Box will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the Box is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the Box will be drawn at a Y coordinate calculated from the formula
"YCoord = ContainerHeight - BoxHeight - YOffset".</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Box *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   value = 0;
   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Boxes[0].Dimensions & DMF_FIXED_YOFFSET) value = Self->Boxes[0].YOffset;
      else if (Self->Boxes[0].Dimensions & DMF_RELATIVE_YOFFSET) {
         value = Self->Boxes[0].YOffset * containerheight / 100;
      }
      else if ((Self->Boxes[0].Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Boxes[0].Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Boxes[0].Dimensions & DMF_FIXED_HEIGHT) height = Self->Boxes[0].Height;
         else height = containerheight * Self->Boxes[0].Height / 100;
         if (Self->Boxes[0].Dimensions & DMF_FIXED_Y) value = containerheight - (Self->Boxes[0].YCoord + height);
         else value = containerheight - ((Self->Boxes[0].YCoord * containerheight / 100) + height);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Box *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->Boxes[0].YOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->Boxes[0].YOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
   else Self->Boxes[0].Dimensions = (Self->Boxes[0].Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
   return(ERR_Okay);
}
