
static void _RedrawDrawable(struct Render *, struct RenderList *, WORD, LONG, LONG, LONG, LONG, struct Bitmap *, LONG);

/*****************************************************************************
** Internal: RedrawNonIntersect()
** Synopsis: void RedrawNonIntersect(RenderID, ClipRectangle *RegionA, ClipRectangle *RegionB, LONG RedrawFlags)
**
** Redraw everything in RegionB that does not intersect with RegionA.
*/

static void RedrawNonIntersect(OBJECTID RenderID, struct ClipRectangle *Region, struct ClipRectangle *RegionB, LONG RedrawFlags, LONG ExposeFlags)
{
   struct ClipRectangle intersect;

   #ifdef DBG_DRAW
      DPrintF("~RedrawNonIntersect:","(A) %dx%d,%dx%d Vs (B) %dx%d,%dx%d", Region->Left, Region->Top, Region->Right, Region->Bottom, RegionB->Left, RegionB->Top, RegionB->Right, RegionB->Bottom);
   #endif

   ExposeFlags |= EXF_ABSOLUTE;

   intersect.left   = RegionB->left;
   intersect.top    = RegionB->top;
   intersect.right  = RegionB->right;
   intersect.bottom = RegionB->bottom;

   if (intersect.right > Region->Right) { /* Right */
      #ifdef DBG_DRAW
         DPrintF("RedrawNonIntersect:","Right exposure");
      #endif

      if (RedrawFlags != -1) RedrawDrawable(RenderID, (intersect.left > Region->Right) ? intersect.left : Region->Right, intersect.top, intersect.right, intersect.bottom, RedrawFlags);
      if (ExposeFlags != -1) ExposeDrawable(RenderID, (intersect.left > Region->Right) ? intersect.left : Region->Right, intersect.top, intersect.right, intersect.bottom, ExposeFlags);
      intersect.right = Region->Right;
      if (intersect.left >= intersect.right) return;
   }

   if (intersect.bottom > Region->Bottom) { /* Bottom */
      #ifdef DBG_DRAW
         DPrintF("RedrawNonIntersect:","Bottom exposure");
      #endif
      if (RedrawFlags != -1) RedrawDrawable(RenderID, intersect.left, (intersect.top > Region->Bottom) ? intersect.top : Region->Bottom, intersect.right, intersect.bottom, RedrawFlags);
      if (ExposeFlags != -1) ExposeDrawable(RenderID, intersect.left, (intersect.top > Region->Bottom) ? intersect.top : Region->Bottom, intersect.right, intersect.bottom, ExposeFlags);
      intersect.bottom = Region->Bottom;
      if (intersect.top >= intersect.bottom) return;
   }

   if (intersect.top < Region->Top) { /* Top */
      #ifdef DBG_DRAW
         DPrintF("RedrawNonIntersect:","Top exposure");
      #endif
      if (RedrawFlags != -1) RedrawDrawable(RenderID, intersect.left, intersect.top, intersect.right, (intersect.bottom < Region->Top) ? intersect.bottom : Region->Top, RedrawFlags);
      if (ExposeFlags != -1) ExposeDrawable(RenderID, intersect.left, intersect.top, intersect.right, (intersect.bottom < Region->Top) ? intersect.bottom : Region->Top, ExposeFlags);
      intersect.top = Region->Top;
   }

   if (intersect.left < Region->Left) { /* Left */
      #ifdef DBG_DRAW
         DPrintF("RedrawNonIntersect:","Left exposure");
      #endif
      if (RedrawFlags != -1) RedrawDrawable(RenderID, intersect.left, intersect.top, (intersect.right < Region->Left) ? intersect.right : Region->Left, intersect.bottom, RedrawFlags);
      if (ExposeFlags != -1) ExposeDrawable(RenderID, intersect.left, intersect.top, (intersect.right < Region->Left) ? intersect.right : Region->Left, intersect.bottom, ExposeFlags);
   }

   #ifdef DBG_DRAW
      StepBack();
   #endif
}

/*****************************************************************************
** Internal: ExposedRegion()
**
** Keep in mind that all coordinates are absolute.
**
** Rect:      The area that you are copying from (NOTE: Fields in this structure will be updated).
** Area:      The clip region.
** Expose:    (Result) The resulting area that lies outside of the clip boundary.
*/

static LONG ExposedRegion(struct ClipRectangle *rect, struct ClipRectangle *Area, struct ClipRectangle *expose)
{
   if (rect->Left < Area->Left) {
      /*** Left edge exposed ***/
      expose->Left = rect->Left;
      expose->Top  = rect->Top;
      if (rect->Right < Area->Left) expose->Right = rect->Right;
      else expose->Right = Area->Left;
      expose->Bottom = rect->Bottom;
      rect->Left = Area->Left;
      return(TRUE);
   }

   if (rect->Top < Area->Top) {
      /*** Top edge exposed ***/
      expose->Left   = rect->Left;
      expose->Top    = rect->Top;
      expose->Right  = rect->Right;
      if (rect->Bottom < Area->Top) expose->Bottom = rect->Bottom;
      else expose->Bottom = Area->Top;
      rect->Top = Area->Top;
      return(TRUE);
   }

   if (rect->Bottom > Area->Bottom) {
      /*** Bottom edge exposed ***/
      expose->Left   = rect->Left;
      if (rect->Top > Area->Bottom) expose->Top = rect->Top;
      else expose->Top = Area->Bottom;
      expose->Right  = rect->Right;
      expose->Bottom = rect->Bottom;
      rect->Bottom = Area->Bottom;
      return(TRUE);
   }

   if (rect->Right > Area->Right) {
      /*** Right edge exposed ***/
      if (rect->Left > Area->Right) expose->Left = rect->Left;
      else expose->Left = Area->Right;
      expose->Right  = rect->Right;
      expose->Top    = rect->Top;
      expose->Bottom = rect->Bottom;
      rect->Right = Area->Right;
      return(TRUE);
   }

   return(FALSE);
}

/*****************************************************************************

###ACTION###
Name:  Draw
Short: Redraws the contents of a render object.

###DESCRIPTION###
<p>Drawing a render object causes all child objects belonging to that render
object to be sent a Draw action.  This has the effect of redrawing all
graphics within the render object.  The procedure is as follows:</p>

<ul>
<li>If the render object's Colour field has been set, it will clear its
background to that colour.</li>
<li>Every child belonging to the render object is then sent a Draw
notification so that the contents are redrawn.</li>
<li>The video display is then updated.</li>
</ul>

<p>There are just two important notes to be aware of: 1. If the render object
contains other render objects, they will not be sent a Draw action unless they
are using special effects such as transparency.  2. If the render object has
not had its background colour set or if masking is not enabled, the current
contents will not be automatically cleared.  (This is advantageous in
situations where a particular object will clear the rendered area first).</p>

###END###

*****************************************************************************/

ERROR RENDER_Draw(struct Render *Self, struct acDraw *Args)
{
   struct ActionMessage *action;
   UBYTE msgbuffer[sizeof(struct Message) + sizeof(struct ActionMessage) + sizeof(struct acDraw)];
   LONG x, y, width, height;
   APTR queue;
   LONG msgindex;

   /*** If the Render object is invisible, return immediately ***/

   if ((!(Self->Flags & RNF_VISIBLE)) OR (glNoDrawing) OR (Self->Width < 1) OR (Self->Height < 1)) {
      #ifdef DBG_DRAW
         DPrintF("Draw:","[Render:%d] Not drawing (invisible or glNoDrawing set).", Self->Head.UniqueID);
      #endif

      return(ERR_Okay|ERF_Notified);
   }

   if (!Args) {
      x = 0;
      y = 0;
      width  = Self->Width;
      height = Self->Height;
   }
   else {
      x      = Args->XCoord;
      y      = Args->YCoord;
      width  = Args->Width;
      height = Args->Height;
      if (!width) width = Self->Width;
      if (!height) height = Self->Height;
   }

   /* Check if other draw messages are queued for this object - if so,
   ** do not do anything until the final message is reached.
   */

   if (AccessMemory(GetResource(RES_MESSAGEQUEUE), MEM_READ, &queue) IS ERR_Okay) {
      msgindex = 0;
      while (ScanMessages(queue, &msgindex, MSGID_ACTION, msgbuffer, sizeof(msgbuffer)) IS ERR_Okay) {
         action = (struct ActionMessage *)(msgbuffer + sizeof(struct Message));

         if ((action->ActionID IS MT_InvalidateRegion) AND (action->ObjectID IS Self->Head.UniqueID)) {
            if (action->SendArgs IS FALSE) {
               ReleaseMemory(queue);
               return(ERR_Okay|ERF_Notified);
            }
         }
         else if ((action->ActionID IS AC_Draw) AND (action->ObjectID IS Self->Head.UniqueID)) {

            if (action->SendArgs IS TRUE) {
               struct acDraw *msgdraw;
               msgdraw = (struct acDraw *)(action + 1);

               if (!Args) {
                  // Tell the next message to draw everything.
                  action->SendArgs = FALSE;
               }
               else {
                  DOUBLE right, bottom;

                  right  = msgdraw->XCoord + msgdraw->Width;
                  bottom = msgdraw->YCoord + msgdraw->Height;

                  if (x < msgdraw->XCoord) msgdraw->XCoord = x;
                  if (y < msgdraw->YCoord) msgdraw->YCoord = y;
                  if ((x + width) > right)   right  = x + width;
                  if ((y + height) > bottom) bottom = y + height;

                  msgdraw->Width  = right - msgdraw->XCoord;
                  msgdraw->Height = bottom - msgdraw->YCoord;
               }

               UpdateMessage(queue, ((struct Message *)msgbuffer)->UniqueID, NULL, action, sizeof(struct ActionMessage) + sizeof(struct acDraw));
            }
            else {
               // We do nothing here because the next draw message will draw everything.
            }

            ReleaseMemory(queue);
            return(ERR_Okay|ERF_Notified);
         }
      }
      ReleaseMemory(queue);
   }

#ifdef DBG_DRAW
   DPrintF("~Draw()","[Render:%d:%s] %dx%d,%dx%d", Self->Head.UniqueID, Self->Head.Stats->Name, x, y, width, height);
#endif

   RedrawDrawable(Self->Head.UniqueID, x, y, width, height, IRF_RELATIVE|IRF_IGNORE_CHILDREN);
   ExposeDrawable(Self->Head.UniqueID, x, y, width, height, EXF_REDRAW_VOLATILE);

#ifdef DBG_DRAW
   StepBack();
#endif

   return(ERR_Okay|ERF_Notified);
}

/*****************************************************************************

###METHOD###
Name:  InvalidateRegion()
Short: Redraws all of the content in a render object.

###DESCRIPTION###
<p>Invalidating a render object will cause everything within a specified area
to be redrawn.  This includes child render objects that intersect with
the area that you have specified.  Parent regions that overlap are not included
in the redraw.</p>

<p>To quickly redraw an entire render object's content, call this method
directly without supplying an argument structure.  If you want to redraw a
render object and ignore all of its render children then you should use the
Draw action instead of this method.</p>

<p>If you want to refresh a rendered area to the display then you should
use the Expose method instead.  Exposing will use the graphics buffer to
refresh the graphics, thus avoiding the speed loss of a complete redraw.</p>

###ARGS###
LONG/XCoord: Horizontal coordinate of the region to invalidate.
LONG/YCoord: Vertical coordinate of the region to invalidate.
LONG/Width:  Width of the region to invalidate.
LONG/Height: Height of the region to invalidate.

###RESULT###
ERR_Okay: Success.
ERR_AccessMemory: Failed to access the internal render list.

###SEE ALSO###
Method: Expose
Action: Draw

###END###

*****************************************************************************/

static ERROR RENDER_InvalidateRegion(struct Render *Self, struct mtInvalidateRegion *Args)
{
   struct ActionMessage *action;
   UBYTE msgbuffer[sizeof(struct Message) + sizeof(struct ActionMessage) + sizeof(struct mtInvalidateRegion)];
   APTR queue;
   LONG msgindex;

   if ((!(Self->Flags & RNF_VISIBLE)) OR (glNoDrawing) OR (Self->Width < 1) OR (Self->Height < 1)) {
      return(ERR_Okay|ERF_Notified);
   }

   /* Check if other draw messages are queued for this object - if so,
   ** do not do anything until the final message is reached.
   */

   if (AccessMemory(GetResource(RES_MESSAGEQUEUE), MEM_READ, &queue) IS ERR_Okay) {
      msgindex = 0;
      while (ScanMessages(queue, &msgindex, MSGID_ACTION, msgbuffer, sizeof(msgbuffer)) IS ERR_Okay) {
         action = (struct ActionMessage *)(msgbuffer + sizeof(struct Message));

         if ((action->ActionID IS MT_InvalidateRegion) AND (action->ObjectID IS Self->Head.UniqueID)) {
            struct mtInvalidateRegion *msginvalid;

            if (action->SendArgs IS TRUE) {
               msginvalid = (struct mtInvalidateRegion *)(action + 1);

               if (!Args) {
                  /*** Invalidate everything ***/
                  action->SendArgs = FALSE;
               }
               else {
                  DOUBLE right, bottom;

                  right  = msginvalid->XCoord + msginvalid->Width;
                  bottom = msginvalid->YCoord + msginvalid->Height;

                  if (Args->XCoord < msginvalid->XCoord) msginvalid->XCoord = Args->XCoord;
                  if (Args->YCoord < msginvalid->YCoord) msginvalid->YCoord = Args->YCoord;
                  if ((Args->XCoord + Args->Width) > right)   right  = Args->XCoord + Args->Width;
                  if ((Args->YCoord + Args->Height) > bottom) bottom = Args->YCoord + Args->Height;

                  msginvalid->Width  = right - msginvalid->XCoord;
                  msginvalid->Height = bottom - msginvalid->YCoord;
               }

               UpdateMessage(queue, ((struct Message *)msgbuffer)->UniqueID, NULL, action, sizeof(struct ActionMessage) + sizeof(struct mtInvalidateRegion));
            }
            else {
               // We do nothing here because the next invalidation message will draw everything.
            }

            ReleaseMemory(queue);
            return(ERR_Okay|ERF_Notified);
         }
      }
      ReleaseMemory(queue);
   }

#ifdef DBG_DRAW
   DPrintF("~InvalidateRegion()","[Render:%d]", Self->Head.UniqueID);
#endif

   if (Args) {
      RedrawDrawable(Self->Head.UniqueID, Args->XCoord, Args->YCoord, Args->Width, Args->Height, IRF_RELATIVE);
      ExposeDrawable(Self->Head.UniqueID, Args->XCoord, Args->YCoord, Args->Width, Args->Height, EXF_CHILDREN|EXF_REDRAW_VOLATILE_OVERLAP);
   }
   else {
      RedrawDrawable(Self->Head.UniqueID, 0, 0, Self->Width, Self->Height, IRF_RELATIVE);
      ExposeDrawable(Self->Head.UniqueID, 0, 0, Self->Width, Self->Height, EXF_CHILDREN|EXF_REDRAW_VOLATILE_OVERLAP);
   }

#ifdef DBG_DRAW
   StepBack();
#endif
   return(ERR_Okay|ERF_Notified);
}

/*****************************************************************************

###PRIVATE###
Name:  RedrawDrawable()
Short: Redraws all of the content in a render object.

###DESCRIPTION###
<p>Invalidating a render object will cause everything within a specified area
to be redrawn.  This includes child render objects that intersect with
the area that you have specified.  Overlapping siblings are not redrawn
unless they are marked as volatile.</p>

<p>To quickly redraw an entire render object's content, call this method
directly without supplying an argument structure.  If you want to redraw a
render object and ignore all of its render children then you should use the
Draw action instead of this method.</p>

<p>If you want to refresh a rendered area to the display then you should use the
ExposeDrawable() function instead.  The ExposeDrawable() function copies the
graphics buffer to the display only, thus avoiding the speed loss of a
complete redraw.</p>

<p>Because RedrawDrawable() only redraws internal graphics buffers, this
function is typically followed with a call to ExposeDrawable().</p>

<p>Flag options:</p>

<pre>
   IRF_IGNORE_CHILDREN   Do not draw child drawables (NB: even volatile children are not drawn).
   IRF_IGNORE_NVCHILDREN Do not draw child drawables unless they are volatile.
   IRF_SINGLE_BITMAP     Only draw children and/or siblings when they share our bitmap space.
   IRF_RELATIVE          Coordinate arguments are expressed in relative format.  Right and Bottom reflect Width and Height respectively.
</pre>

###INPUT###
OBJECTID/RenderID: The ID of the drawable that you want to invalidate.
LONG/Left: Absolute horizontal coordinate of the region to invalidate.
LONG/Top: Absolute vertical coordinate of the region to invalidate.
LONG/Right:  Absolute right-hand coordinate of the region to invalidate.
LONG/Bottom: Absolute bottom coordinate of the region to invalidate.
LONG/Flags: Optional flags.

###RESULT###
ERR_Okay: Success.
ERR_AccessMemory: Failed to access the internal render list.

###SEE ALSO###
Function: ExposeDrawable()

###END###

*****************************************************************************/

static ERROR RedrawDrawable(OBJECTID RenderID, LONG Left, LONG Top, LONG Right, LONG Bottom, LONG Flags)
{
   static BYTE recursive = 0;
   struct RenderList *list;
   struct Bitmap *bitmap;
   struct Render *render;
   struct MemInfo meminfo;
   OBJECTID parent_id;
   WORD level, index, i;
   ERROR error;

   if (glNoDrawing) return(ERR_Okay);

   if (!(list = AccessRenderList(ARF_READ, NULL))) {
      DPrintF("@Redraw()","Unable to access the renderlist.");
      return(ERR_AccessMemory);
   }

   if ((index = FindDrawableIndex(list, RenderID)) IS -1) goto exit;

   if (list[index].Flags & RNF_TOTALREDRAW) {
      /* If the TOTALREDRAW flag is set against the drawable then the entire drawable must be
      ** redrawn regardless of the circumstances.  This is often required for algorithmic effects
      ** as seen in the Blur class.
      */
      Left   = list[index].Left;
      Top    = list[index].Top;
      Right  = list[index].Right;
      Bottom = list[index].Bottom;
   }
   else if (Flags & IRF_RELATIVE) {
      Left   = list[index].Left + Left;
      Top    = list[index].Top + Top;
      Right  = Left + Right;
      Bottom = Top + Bottom;
      Flags &= ~IRF_RELATIVE;
   }

#ifdef DBG_DRAW
   DPrintF("~Redraw()","[%d] Size: %dx%d,%dx%d Expose: %dx%d,%dx%d $%.4x", RenderID, list[index].AbsX, list[index].AbsY, list[index].Width, list[index].Height, Left, Top, Right-Left, Bottom-Top, Flags);
#endif

   if ((list[index].Flags & (RNF_REGION|RNF_INVISIBLE)) AND (!recursive)) {
      #ifdef DBG_DRAW
         DPrintF("Redraw:","Passing draw request to parent (I am a %s)", (list[index].Flags & RNF_REGION) ? "region" : "invisible");
      #endif
      RedrawDrawable(list[index].ParentID, Left, Top, Right, Bottom, Flags & (~IRF_IGNORE_CHILDREN));
      goto exit;
   }

   /*** Check if any of the parent drawables are invisible ***/

   if ((!(list[index].Flags & RNF_VISIBLE)) OR (CheckVisibility(list, index) IS FALSE)) {
      #ifdef DBG_DRAW
         DPrintF("Redraw:","Drawable is not visible.");
      #endif
      goto exit;
   }

   /* Because we are executing a redraw, we need to ensure that the drawable
   ** is within our task space before going any further.  The best way to do this
   ** is through MemoryIDInfo() so that we don't risk locking up the system via
   ** AccessObject().
   */

   if (MemoryIDInfo(list[index].ObjectID, VER_MemoryInfo, &meminfo) IS ERR_Okay) {
      if (meminfo.TaskID != CurrentTaskID()) {
         #ifdef DBG_DRAW
            DPrintF("@Redraw:","Render object #%d belongs to task #%d (we are #%d)", RenderID, meminfo.TaskID, CurrentTaskID());
         #endif

         if (Flags & IRF_IGNORE_CHILDREN) {
            struct acDraw draw;
            draw.XCoord = Left - list[index].AbsX;
            draw.YCoord = Top - list[index].AbsY;
            draw.Width  = Right - Left;
            draw.Height = Bottom - Top;
            ActionMsg(AC_Draw, list[index].ObjectID, &draw);
         }
         else {
            struct mtInvalidateRegion invalid;
            invalid.XCoord = Left - list[index].AbsX;
            invalid.YCoord = Top - list[index].AbsY;
            invalid.Width  = Right - Left;
            invalid.Height = Bottom - Top;
            ActionMsg(MT_InvalidateRegion, list[index].ObjectID, &invalid);
         }
         goto exit;
      }
   }
   else goto exit;

   /* Check if the exposed dimensions are outside of our boundary and/or
   ** our parent(s) boundaries.  If so then we must restrict the exposed
   ** dimensions.
   */

   parent_id = RenderID;
   i = index;
   while (parent_id) {
      while ((list[i].ObjectID != parent_id) AND (i > 0)) i--;

      if (list[i].BitmapID != list[index].BitmapID) break; /* Stop if we encounter a separate bitmap */

      if (Left   < list[i].Left)   Left   = list[i].Left;
      if (Top    < list[i].Top)    Top    = list[i].Top;
      if (Right  > list[i].Right)  Right  = list[i].Right;
      if (Bottom > list[i].Bottom) Bottom = list[i].Bottom;

      parent_id = list[i].ParentID;
   }

   if ((Left >= Right) OR (Top >= Bottom)) goto exit;

   /*** Draw the render graphics into the bitmap buffer ***/

   if ((error = AccessObject(list[index].ObjectID, 5000, &render)) IS ERR_Okay) {
      #ifdef DBG_DRAW
         DPrintF("Redraw:","Expose: %dx%d,%dx%d", Left, Top, Right-Left, Bottom-Top);
      #endif

      if (AccessObject(list[index].BitmapID, 5000, &bitmap) IS ERR_Okay) {
         /* First, check if there has been a change in the video bit depth.  If so,
         ** regenerate the bitmap with a matching depth.
         */

         CheckBitmapBufferDepth(render, bitmap);
         _RedrawDrawable(render, list, index, Left, Top, Right, Bottom, bitmap, (Flags & (IRF_IGNORE_CHILDREN|IRF_IGNORE_NVCHILDREN)) ? 0 : URF_HATE_CHILDREN);
         ReleaseObject(bitmap);
      }
      else {
         DPrintF("@Redraw:","Unable to access bitmap object #%d.", list[index].BitmapID);
         ReleaseObject(render);
         goto exit;
      }

      ReleaseObject(render);

      /* We have done the redraw, so now we can send invalidation messages
      ** to any intersecting -child- drawables for this region.  This process is
      ** not recursive (notice the use of IRF_IGNORE_CHILDREN) but all children
      ** will be covered due to the way the tree is traversed.
      */

      if (!(Flags & IRF_IGNORE_CHILDREN)) {
         level = list[index].Level;
         for (i=index+1; list[i].ObjectID; i++) {
            if (list[i].Level <= level) break; /* End of list - exit this loop */

            if (Flags & IRF_IGNORE_NVCHILDREN) {
               /* Ignore children except for those that are volatile */
               if (!(list[i].Flags & RNF_VOLATILE)) continue;
            }
            else {
               if ((Flags & IRF_SINGLE_BITMAP) AND (list[i].BitmapID != list[index].BitmapID)) continue;
            }

            if ((list[i].Flags & (RNF_REGION|RNF_CURSOR)) OR (!(list[i].Flags & RNF_VISIBLE))) {
               continue; /* Skip regions and non-visible children */
            }

            if ((list[i].Right > Left) AND (list[i].Bottom > Top) AND
                (list[i].Left < Right) AND (list[i].Top < Bottom)) {
               recursive++;
               RedrawDrawable(list[i].ObjectID, Left, Top, Right, Bottom, Flags|IRF_IGNORE_CHILDREN);
               recursive--;
            }
         }
      }
   }
   else {
      /* If the object does not exist then its task has crashed and
      ** we need to remove it from the render list.
      */

      if (error IS ERR_NoMatchingObject) {
         DPrintF("@Redraw:","Removing references to render object #%d (owner crashed).", list[index].ObjectID);
         UntrackLayer(list[index].ObjectID);
      }
      else DPrintF("@Redraw:","Unable to access render object #%d, error %d.", list[index].ObjectID, error);
   }

exit:
   ReleaseRenderList(ARF_READ);

#ifdef DBG_DRAW
   StepBack();
#endif
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: _RedrawDrawable()
** Short:    This function fulfills the recursive drawing requirements of
**           RedrawDrawable().  This function is not intended for any other
**           use.
*/

static void _RedrawDrawable(struct Render *Self, struct RenderList *list, WORD Index,
                         LONG Left, LONG Top, LONG Right, LONG Bottom, struct Bitmap *DestBitmap, LONG Flags)
{
   struct ClipRectangle abs;
   WORD i, level;
   LONG listx, listy, listright, listbottom;

   if (Self->Flags & (RNF_REGION|RNF_INVISIBLE)) return;

   abs.left   = Left;
   abs.top    = Top;
   abs.right  = Right;
   abs.bottom = Bottom;
   if (abs.left   < list[Index].AbsX)   abs.left   = list[Index].Left;
   if (abs.top    < list[Index].AbsY)   abs.top    = list[Index].Top;
   if (abs.right  > list[Index].Right)  abs.right  = list[Index].Right;
   if (abs.bottom > list[Index].Bottom) abs.bottom = list[Index].Bottom;

/*
   if (Left   < list[Index].AbsX)   Left   = list[Index].Left;
   if (Top    < list[Index].AbsY)   Top    = list[Index].Top;
   if (Right  > list[Index].Right)  Right  = list[Index].Right;
   if (Bottom > list[Index].Bottom) Bottom = list[Index].Bottom;
*/
   level = list[Index].Level + 1;   /* The +1 is used to include children contained in the render object */

   for (i=Index+1; list[i].Level > 1; i++) {
      if (list[i].Level < level) level = list[i].Level;

      /*** If the listed object obscures our render area, analyse the region around it ***/

      if (list[i].Level <= level) {
         /* If we have a bitmap buffer and the underlying child region also has its own bitmap,
         ** we have to ignore it in order for our graphics buffer to be correct when
         ** exposes are made.
         */

         if (list[i].BitmapID != Self->BufferID) continue;

         if (!(list[i].Flags & RNF_VISIBLE)) continue;

         if (list[i].Flags & RNF_REGION) continue; /* Regions are completely ignored because it is impossible for them to contain true render layers */

         /*** Check for an intersection and respond to it ***/

         listx      = list[i].Left;
         listy      = list[i].Top;
         listright  = list[i].Right;
         listbottom = list[i].Bottom;

         if ((listx < Right) AND (listy < Bottom) AND (listright > Left) AND (listbottom > Top)) {
            if (list[i].Flags & RNF_CURSOR) {
               /* Objects like the pointer cursor are ignored completely.  They
               ** are redrawn following exposure.
               */
               return;
            }
            else if (list[i].Flags & RNF_INVISIBLE) {
               /* If the render object is see-through then we will ignore its bounds, but legally
               ** it can also contain child render objects that are solid.  For that reason,
               ** we have to 'go inside' to check for solid children and draw around them.
               */

               _RedrawDrawable(Self, list, i, Left, Top, Right, Bottom, DestBitmap, Flags);

               return;
            }

            if (Flags & URF_HATE_CHILDREN) {
               /* The HATE_CHILDREN flag is used if the caller intends to redraw all children drawables.
               ** In this case, we may as well ignore children when they are smaller than 100x100 in size,
               ** because splitting our drawing process into four sectors is probably going to be slower
               ** than just redrawing the entire background in one shot.
               */
               if (list[i].Width + list[i].Height <= 200) continue;
            }

            if (listx <= Left) listx = Left;
            else _RedrawDrawable(Self, list, Index, Left, Top, listx, Bottom, DestBitmap, Flags); /* left */

            if (listright >= Right) listright = Right;
            else _RedrawDrawable(Self, list, Index, listright, Top, Right, Bottom, DestBitmap, Flags); /* right */

            if (listy <= Top) listy = Top;
            else _RedrawDrawable(Self, list, Index, listx, Top, listright, listy, DestBitmap, Flags); /* top */

            if (listbottom < Bottom) _RedrawDrawable(Self, list, Index, listx, listbottom, listright, Bottom, DestBitmap, Flags); /* bottom */

            return;
         }
      }
   }

#ifdef DBG_DRAW
   DPrintF("~RedrawDrawable:","[%d:%s] %dx,%dy,%dw,%dh", Self->Head.UniqueID, Self->Head.Stats->Name, Left, Top, Right-Left, Bottom-Top);
#endif

   /* If we have been called recursively due to the presence of volatile/invisible regions (see above),
   ** our Index field will not match with the drawable that is referenced in Self.  We need to ensure
   ** correctness before going any further.
   */

   if (list[Index].ObjectID != Self->Head.UniqueID) {
      for (Index=0; (list[Index].ObjectID) AND (list[Index].ObjectID != Self->Head.UniqueID); Index++);
   }

   /*** Prepare the buffer so that it matches the exposed area ***/

   if (Self->prvBitmapOwnerID != Self->Head.UniqueID) {
      for (i=Index; (i > 0) AND (list[i].ObjectID != Self->prvBitmapOwnerID); i--);
      DestBitmap->XOffset = list[Index].AbsX - list[i].AbsX; /* Offset is relative to the bitmap owner */
      DestBitmap->YOffset = list[Index].AbsY - list[i].AbsY;

   }
   else {
      /*** Set the clipping so that we only draw the area that has been exposed ***/
      DestBitmap->XOffset = 0;
      DestBitmap->YOffset = 0;
   }

   DestBitmap->ClipLeft   = Left - list[Index].Left;
   DestBitmap->ClipTop    = Top - list[Index].Top;
   DestBitmap->ClipRight  = Right - list[Index].Left;
   DestBitmap->ClipBottom = Bottom - list[Index].Top;

   // THIS SHOULD NOT BE NEEDED - but occasionally it detects render problems (bugs in other areas of the render code?)

   if (((DestBitmap->XOffset + DestBitmap->ClipLeft) < 0) OR ((DestBitmap->YOffset + DestBitmap->ClipTop) < 0) OR
       ((DestBitmap->XOffset + DestBitmap->ClipRight) > DestBitmap->Width) OR ((DestBitmap->YOffset + DestBitmap->ClipBottom) > DestBitmap->Height)) {
      DPrintF("@UpdateRegion()","Invalid coordinates detected (outside of the drawable area).");
      if ((DestBitmap->XOffset + DestBitmap->ClipLeft) < 0) DestBitmap->ClipLeft = 0;
      if ((DestBitmap->YOffset + DestBitmap->ClipTop) < 0)  DestBitmap->ClipTop = 0;
      DestBitmap->ClipRight = DestBitmap->Width - DestBitmap->XOffset;
      DestBitmap->ClipBottom = DestBitmap->Height - DestBitmap->YOffset;
   }

   /*** Clear the background ***/

   if ((Self->Flags & RNF_PRECOPY) AND (!(Self->Flags & RNF_POSTCOMPOSITE))) {
      PrepareBackground(Self, list, Index, DestBitmap, &abs, STAGE_PRECOPY);
   }
   else if (Self->Flags & RNF_POSTCOMPOSITE) {
      ActionTags(MT_DrawRectangle, DestBitmap, 0, 0, Self->Width, Self->Height,
         PackPixelA((struct Bitmap *)DestBitmap, 0, 0, 0, 255), TRUE);
   }
   else if (Self->Flags & RNF_CLEAR) {
      ActionTags(MT_DrawRectangle, DestBitmap, 0, 0, Self->Width, Self->Height,
         PackPixel((struct Bitmap *)DestBitmap, Self->prvColourRGB.Red, Self->prvColourRGB.Green, Self->prvColourRGB.Blue), TRUE);
   }

   /*** Draw graphics to the buffer ***/

   glFreeExpose = DestBitmap->Head.UniqueID;

      /*** Call callbacks ***/

      if (Self->prvCallbackCount <= ARRAYSIZE(Self->prvCallbackCache)) {
         /*** Ensure that the prvCallback address is valid ***/
         Self->prvCallback = Self->prvCallbackCache;
      }

      for (i=0; i < Self->prvCallbackCount; i++) {
         void (*routine)(LONG, struct Render *, struct Bitmap *);

         routine = Self->prvCallback[i].Routine;

#ifdef DBG_DRAW_ROUTINES
         DPrintF("~RedrawDrawable:","Callback %d: Routine: $%.8x", i, routine);
#endif
         routine(Self->prvCallback[i].Data, Self, DestBitmap);

#ifdef DBG_DRAW_ROUTINES
         StepBack();
#endif
      }

      /*** Call notification subscribers ***/

      Self->prvDrawableBitmap = DestBitmap;
      NotifySubscribers(Self, AC_Draw, NULL, NSF_LOCALTASK);
      Self->prvDrawableBitmap = NULL;

   glFreeExpose = NULL;

   /*** After-copy management ***/

   if (!(Self->Flags & RNF_POSTCOMPOSITE)) {
      if (Self->Flags & RNF_AFTERCOPY) {
         PrepareBackground(Self, list, Index, DestBitmap, &abs, STAGE_AFTERCOPY);
      }
      else if (Self->Flags & RNF_ROOTLAYER) {
         /* If the render object is part of a global background, we have to look
         ** for the root layer and check if it has the AFTERCOPY flag set.
         */
         for (i=0; list[i].ObjectID; i++) {
            if (list[i].ObjectID IS Self->RootLayer) {
               if (list[i].Flags & RNF_AFTERCOPY) {
                  PrepareBackground(Self, list, Index, DestBitmap, &abs, STAGE_AFTERCOPY);
               }
               break;
            }
         }
      }
   }

   #ifdef DBG_DRAW
      StepBack();
   #endif
}

/*****************************************************************************
** Internal: PrepareBackground()
** Short:    Used for PRECOPY drawables.
**
** Index: The index of our drawable.
** Clip:  The absolute coordinates of the expose area.
*/

static void PrepareBackground(struct Render *Self, struct RenderList *list, WORD Index,
   struct Bitmap *DestBitmap, struct ClipRectangle *clip, BYTE Stage)
{
   struct ClipRectangle expose;
   OBJECTID id;
   WORD i, j, end, master, parentindex, opaque;
   BYTE pervasive;

   #ifdef DBG_DRAW
      DPrintF("~PrepareBkgd()","%d, Position: %dx,%dy %dw,%dh", list[Index].ObjectID, clip->left, clip->top, clip->right-clip->left, clip->bottom-clip->top);
   #endif

   end = Index;
   master = Index;

   if ((Self) AND (list[end].ObjectID != Self->RootLayer)) {
      for (j=0; list[j].ObjectID; j++) {
         if (list[j].ObjectID IS Self->RootLayer) {
            /* Root layers are only considered when they are volatile (otherwise the render object's
            ** current opacity settings take precedence).  This ensures that objects like translucent
            ** scrollbars can take priority if the parent is not translucent.
            **
            ** If a custom root layer has been specified, then we are forced into using it.
            */

            if (list[Index].Flags & RNF_CUSTOMROOT) end = j;
            else if (list[j].Flags & RNF_VOLATILE) end = j;
            break;
         }
      }
   }

   /* Find the parent that owns this drawable (we will use this as the starting point
   ** for our copy operation).  Everything that gets in the way between the parent and
   ** the location of our drawable is what will be copied across.
   */

   parentindex = end;
   while ((parentindex > 0) AND (list[parentindex].ObjectID != list[end].ParentID)) parentindex--;

   /*** If the parent object is invisible, we need to scan back to a visible parent ***/

   id = list[parentindex].ObjectID;
   for (j=parentindex; list[parentindex].Level > 1; j--) {
      if (list[j].ObjectID IS id) {
         if (!(list[j].Flags & RNF_INVISIBLE)) break;
         id = list[j].ParentID;
      }
   }
   parentindex = j;

   /* This loop will scan everything in backwards fashion for copying to the
   ** buffered graphics area.
   */

   for (i=end-1; i >= parentindex; i--) {
      /*** Ignore regions ***/

      if (list[i].Flags & (RNF_REGION|RNF_INVISIBLE|RNF_CURSOR)) continue;

      /*** Take a copy of the expose coordinates ***/

      expose.left   = clip->left;
      expose.top    = clip->top;
      expose.right  = clip->right;
      expose.bottom = clip->bottom;

      /*** Check the visibility of this layer and its parents ***/

      id = list[i].ObjectID;
      for (j=i; (j >= 0) AND (id); j--) {
         if (list[j].ObjectID IS id) {
            if (!(list[j].Flags & RNF_VISIBLE)) goto next;

            id = list[j].ParentID;

            if (list[j].BitmapID IS list[i].BitmapID) {
               if (expose.left   < list[j].Left)   expose.left   = list[j].AbsX;
               if (expose.top    < list[j].Top)    expose.top    = list[j].AbsY;
               if (expose.right  > list[j].Right)  expose.right  = list[j].Right;
               if (expose.bottom > list[j].Bottom) expose.bottom = list[j].Bottom;
            }
         }
      }

      if ((expose.right <= expose.left) OR (expose.bottom <= expose.top)) continue;

      /*** Do the expose ***/

#ifdef FASTACCESS
      if (Stage IS STAGE_AFTERCOPY) {
         if (list[Index].Flags & RNF_CUSTOMROOT) opaque = list[Index].Opacity;
         else opaque = list[end].Opacity;
      }
      else opaque = 255;

      if ((list[Index].Flags & RNF_PERVASIVECOPY) AND (Stage IS STAGE_AFTERCOPY)) pervasive = TRUE;
      else pervasive = FALSE;

      FastCopyBackground(list, i, end, master, expose.left, expose.top, expose.right, expose.bottom, DestBitmap, list[i].ObjectID, opaque, pervasive);
#else
{
      struct Bitmap *bitmap;

      if (AccessObject(list[i].BitmapID, 2000, &bitmap) IS ERR_Okay) {
         if (Stage IS STAGE_AFTERCOPY) {
            if (list[Index].Flags & RNF_CUSTOMROOT) opaque = list[Index].Opacity;
            else opaque = list[end].Opacity;
         }
         else opaque = 255;

         if ((list[Index].Flags & RNF_PERVASIVECOPY) AND (Stage IS STAGE_AFTERCOPY)) pervasive = TRUE;
         else pervasive = FALSE;

         CopyBackground(list, i, end, master, expose.left, expose.top, expose.right, expose.bottom, DestBitmap, bitmap, opaque, pervasive);
         ReleaseObject(bitmap);
      }
      else {
         DPrintF("@PrepareBkgd:","[Render:%d] Failed to access bitmap #%d.", list[Index].ObjectID, list[i].BitmapID);
         break;
      }
}
#endif
next:
      continue;
   }

   #ifdef DBG_DRAW
      StepBack();
   #endif
}

/*****************************************************************************
** Internal: FastCopyBackground()
*/

#ifdef FASTACCESS
static void FastCopyBackground(struct RenderList *list, WORD Index, WORD End, WORD Master,
   WORD Left, WORD Top, WORD Right, WORD Bottom, struct Bitmap *DestBitmap,
   OBJECTID SourceID, WORD Opacity, BYTE Pervasive)
{
   WORD i, j, listx, listy, listright, listbottom, root;
   struct BitmapSurface surface;
   LONG flags;

   /*** Scan for overlapping parent/sibling regions and avoid them ***/

   for (i=Index+1; (i < End) AND (list[i].Level > 1); i++) {

      /*** Skip past non-visible areas and their content ***/

      if (!(list[i].Flags & RNF_VISIBLE)) {
         j = list[i].Level;
         while (list[i+1].Level > j) i++;
         continue;
      }

      if (list[i].Flags & (RNF_REGION|RNF_INVISIBLE|RNF_CURSOR)) continue; /* Ignore regions */

      /*** If the copy is pervasive then all children must be ignored (so that we can blit translucent graphics over them) ***/

      if ((Pervasive IS TRUE) AND (list[i].Level > list[Index].Level)) continue;

      //if (list[i].BitmapID IS list[Index].BitmapID) continue;

      listx      = list[i].AbsX;
      listy      = list[i].AbsY;
      listright  = list[i].Right;
      listbottom = list[i].Bottom;

      if ((listx < Right) AND (listy < Bottom) AND (listright > Left) AND (listbottom > Top)) {
         if (listx <= Left) listx = Left;
         else FastCopyBackground(list, Index, End, Master, Left, Top, listx, Bottom, DestBitmap, SourceID, Opacity, Pervasive); /* left */

         if (listright >= Right) listright = Right;
         else FastCopyBackground(list, Index, End, Master, listright, Top, Right, Bottom, DestBitmap, SourceID, Opacity, Pervasive); /* right */

         if (listy <= Top) listy = Top;
         else FastCopyBackground(list, Index, End, Master, listx, Top, listright, listy, DestBitmap, SourceID, Opacity, Pervasive); /* top */

         if (listbottom < Bottom) FastCopyBackground(list, Index, End, Master, listx, listbottom, listright, Bottom, DestBitmap, SourceID, Opacity, Pervasive); /* bottom */

         return;
      }

      /* Skip past any children of the overlapping object.  This ensures that
      ** we only look at immediate parents and siblings that are in our way.
      */

      j = i + 1;
      while (list[j].Level > list[i].Level) j++;
      i = j - 1;
   }

   /* Check if the exposed dimensions are outside of our boundary and/or
   ** our parent(s) boundaries.  If so then we must restrict the exposed
   ** dimensions.
   */

   parent_id = list[Index].ObjectID;
   for (i=Index; i >= 0;) {
      if (Left   < list[i].AbsX) Left = list[i].AbsX;
      if (Top    < list[i].AbsY) Top  = list[i].AbsY;
      if (Right  > list[i].Right)  Right  = list[i].Right;
      if (Bottom > list[i].Bottom) Bottom = list[i].Bottom;
      parent_id = list[i--].ParentID;
      while ((i > 0) AND (list[i].ObjectID != parent_id)) i--;
   }

   if ((Left >= Right) OR (Top >= Bottom)) goto exit;

   #ifdef DBG_DRAW
      DPrintF("~FastCopyBkgd()","%d, Pos: %dx,%dy %dw,%dh, Bitmap: %d, Index: %d", list[Index].ObjectID, Left, Top, Right - Left, Bottom - Top, list[Index].BitmapID, Index);
   #endif

   /*** The region is not obscured, so perform the redraw ***/

   root = FindBitmapOwner(list, Index);

   if (AccessMemory(list[root].DataMID, MEM_READ, &surface.Data) IS ERR_Okay) {
      flags = BSF_DEFAULTFORMAT;

      surface.ByteWidth     = list[root].LineWidth;
      surface.Height        = list[Index].Height;
      surface.BitsPerPixel  = list[root].BitsPerPixel;
      surface.BytesPerPixel = list[root].BytesPerPixel;
      if (Opacity < 255) {
         surface.Opacity = 255 - Opacity;
         flags |= BSF_TRANSLUCENT;
      }

      BlitSurface(&surface, DestBitmap, flags,
         Left - list[root].AbsX,
         Top - list[root].AbsY,
         Right - Left,
         Bottom - Top,
         Left - list[Master].AbsX,
         Top - list[Master].AbsY);

      ReleaseMemory(surface.Data);
   }

   #ifdef DBG_DRAW
      StepBack();
   #endif
}

#else

static void CopyBackground(struct RenderList *list, WORD Index, WORD End, WORD Master,
   WORD Left, WORD Top, WORD Right, WORD Bottom, struct Bitmap *DestBitmap,
   struct Bitmap *SrcBitmap, WORD Opacity, BYTE Pervasive)
{
   struct mtBlitArea blit;
   WORD i, j, owner, listx, listy, listright, listbottom;
   OBJECTID parent_id;

   /*** Scan for overlapping parent/sibling regions and avoid them ***/

   for (i=Index+1; (i < End) AND (list[i].Level > 1); i++) {
      if (list[i].Flags & (RNF_REGION|RNF_CURSOR)) goto skip; /* Ignore regions */
      if (!(list[i].Flags & RNF_VISIBLE)) goto skip; /* Skip hidden drawables and their content */
      //if (CheckVisibility(list, i) IS FALSE) goto skip;

      if (list[i].Flags & RNF_INVISIBLE) continue; /* Invisibles may contain important regions we have to block */

      /*** If the copy is pervasive then all children must be ignored (so that we can blit translucent graphics over them) ***/

      if ((Pervasive IS TRUE) AND (list[i].Level > list[Index].Level)) goto skip;

      //if (list[i].BitmapID IS list[Index].BitmapID) continue;

      listx      = list[i].Left;
      listy      = list[i].Top;
      listright  = list[i].Right;
      listbottom = list[i].Bottom;

      if ((listx < Right) AND (listy < Bottom) AND (listright > Left) AND (listbottom > Top)) {
         if (listx <= Left) listx = Left;
         else CopyBackground(list, Index, End, Master, Left, Top, listx, Bottom, DestBitmap, SrcBitmap, Opacity, Pervasive); /* left */

         if (listright >= Right) listright = Right;
         else CopyBackground(list, Index, End, Master, listright, Top, Right, Bottom, DestBitmap, SrcBitmap, Opacity, Pervasive); /* right */

         if (listy <= Top) listy = Top;
         else CopyBackground(list, Index, End, Master, listx, Top, listright, listy, DestBitmap, SrcBitmap, Opacity, Pervasive); /* top */

         if (listbottom < Bottom) CopyBackground(list, Index, End, Master, listx, listbottom, listright, Bottom, DestBitmap, SrcBitmap, Opacity, Pervasive); /* bottom */

         return;
      }

      /* Skip past any children of the overlapping object.  This ensures that
      ** we only look at immediate parents and siblings that are in our way.
      */
skip:
      j = i + 1;
      while (list[j].Level > list[i].Level) j++;
      i = j - 1;
   }

   /* Check if the exposed dimensions are outside of our boundary and/or
   ** our parent(s) boundaries.  If so then we must restrict the exposed
   ** dimensions.
   */

   parent_id = list[Index].ObjectID;
   for (i=Index; i >= 0;) {
      if (Left   < list[i].Left)   Left   = list[i].Left;
      if (Top    < list[i].Top)    Top    = list[i].Top;
      if (Right  > list[i].Right)  Right  = list[i].Right;
      if (Bottom > list[i].Bottom) Bottom = list[i].Bottom;
      parent_id = list[i--].ParentID;
      while ((i > 0) AND (list[i].ObjectID != parent_id)) i--;
   }

   if ((Left >= Right) OR (Top >= Bottom)) return;

   #ifdef DBG_DRAW
      DPrintF("~CopyBkgd()","[%d] Pos: %dx%d,%dx%d Bitmap: %d, Index: %d/%d", list[Index].ObjectID, Left, Top, Right - Left, Bottom - Top, list[Index].BitmapID, Index, End);
   #endif

   /*** The region is not obscured, so perform the redraw ***/

   owner = FindBitmapOwner(list, Index);
   SrcBitmap->XOffset = list[Index].Left - list[owner].Left; /* Offset is relative to the bitmap owner */
   SrcBitmap->YOffset = list[Index].Top  - list[owner].Top;

   SrcBitmap->ClipLeft   = Left   - list[Index].Left;
   SrcBitmap->ClipTop    = Top    - list[Index].Top;
   SrcBitmap->ClipRight  = Right  - list[Index].Left;
   SrcBitmap->ClipBottom = Bottom - list[Index].Top;
   if (SrcBitmap->ClipRight  > SrcBitmap->Width)  SrcBitmap->ClipRight  = SrcBitmap->Width;
   if (SrcBitmap->ClipBottom > SrcBitmap->Height) SrcBitmap->ClipBottom = SrcBitmap->Height;

   //DPrintF("Source:","Offset: %d,%d Clip: %d-%dh, %d-%dv", SrcBitmap->XOffset, SrcBitmap->YOffset, SrcBitmap->ClipLeft, SrcBitmap->ClipRight, SrcBitmap->ClipTop, SrcBitmap->ClipBottom);

   if (Opacity < 255) SrcBitmap->Opacity = 255 - Opacity;

   blit.DestBitmap = DestBitmap;
   blit.Flags  = NULL;
   blit.XCoord = SrcBitmap->ClipLeft;
   blit.YCoord = SrcBitmap->ClipTop;
   blit.Width  = SrcBitmap->ClipRight  - SrcBitmap->ClipLeft;
   blit.Height = SrcBitmap->ClipBottom - SrcBitmap->ClipTop;
   blit.XDest  = Left - list[Master].Left;
   blit.YDest  = Top  - list[Master].Top;
   //DPrintF("Blit:","%dx%d TO %dx%d, Size: %dx%d", blit.XCoord, blit.YCoord, blit.XDest, blit.YDest, blit.Width, blit.Height);
   Action(MT_BlitArea, SrcBitmap, &blit);

   SrcBitmap->Opacity = 255;

   #ifdef DBG_DRAW
      StepBack();
   #endif
}

#endif

/*****************************************************************************
** Internal: AccessVideo()
*/

#ifdef _WIN32
static APTR glWindowHandle = NULL;
#else
static LONG glWindowHandle = NULL;
#endif

static struct Bitmap * AccessVideo(OBJECTID ScreenID)
{
   ERROR error;

   if (glVideoCount) {
      glVideoCount++;
      return(glVideoBitmap);
   }

   if ((error = AccessObject(ScreenID, 5000, &glScreen)) IS ERR_Okay) {
      glVideoCount++;
      glVideoBitmap = glScreen->Bitmap;

      /* Because this is a video bitmap, the drawable area has to refer to the
      ** display, which is wholly dependent on the target platform.
      */

      if (GetField(glScreen, FID_WindowHandle, FT_LONG, &glWindowHandle) IS ERR_Okay) {
         #ifdef _WIN32
            SetFields(glVideoBitmap, FID_Drawable|TLONG, winGetDC(glWindowHandle), TAGEND);
         #else
            SetFields(glVideoBitmap, FID_Drawable|TLONG, glWindowHandle, TAGEND);
         #endif
      }

      return(glVideoBitmap);
   }
   else {
      DPrintF("@AccessVideo()","[Render] Unable to gain access to screen #%d, error %d.", ScreenID, error);
      return(NULL);
   }
}

/*****************************************************************************
** Internal: ReleaseVideo()
*/

static void ReleaseVideo(void)
{
   glVideoCount--;

   if (!glVideoCount) {

      #ifdef _WIN32
         APTR drawable;
         GetField(glVideoBitmap, FID_Drawable, FT_LONG, &drawable);
         winReleaseDC(glWindowHandle, drawable);
      #endif

      acFlush(glScreen);

      ReleaseObject(glScreen);

      glScreen       = NULL;
      glVideoBitmap  = NULL;
   }
}

/******************************************************************************

******************************************************************************/

static void CallbackRegion(struct Render *Self, struct RenderList *list, WORD Index,
                         WORD AbsX, WORD AbsY, WORD Right, WORD Bottom, struct Bitmap *DestBitmap);

ERROR ClippedCallback(struct Render *Self, LONG XCoord, LONG YCoord, LONG Width, LONG Height)
{
   struct acResize resize;
   SCREENINFO info;
   struct Bitmap *bitmap;
   struct RenderList *list;
   OBJECTID ParentID;
   WORD i, parentx, parenty, absx, absy;
   WORD parentwidth, parentheight, index, parentindex;

   if ((Self->Width < 1) OR (Self->Height < 1)) return(ERR_Okay);

   /*** Determine the position and size of the area that is going to be drawn ***/

   if (!Width) Width = Self->Width;
   else if (Width > 16384) Width = 16384;

   if (!Height) Height = Self->Height;
   else if (Height > 16384) Height = 16384;

   if ((Width < 1) OR (Height < 1)) {
      DPrintF("@Draw()","Bad arguments - Width: %.2f, Height %.2f", Width, Height);
      return(ERR_Args);
   }

   /*** Restrict the given dimensions to fit inside our rendered area ***/

   if ((XCoord >= Self->Width) OR (YCoord >= Self->Height)) return(ERR_Okay);
   if ((XCoord + Width < 1) OR (YCoord + Height < 1)) return(ERR_Okay);

   if (XCoord < 0) { Width += XCoord; XCoord = 0; }
   if (YCoord < 0) { Height += YCoord; YCoord = 0; }
   if (XCoord + Width > Self->Width) Width = Self->Width - XCoord;
   if (YCoord + Height > Self->Height) Height = Self->Height - YCoord;

   /* Access the render list and perform some more clipping
   ** calculations based on the parent, then perform the draw.
   */

   if ((list = AccessRenderList(ARF_READ, NULL))) {

      /*** Find our object ***/

      for (index=0; (list[index].ObjectID) AND (list[index].ObjectID != Self->Head.UniqueID); index++);

      if (list[index].ObjectID != Self->Head.UniqueID) {
         ReleaseRenderList(ARF_READ);
         return(SendError(ERH_Function, ERR_Search));
      }

      absx = list[index].AbsX;
      absy = list[index].AbsY;

      /* Check if the exposed dimensions are outside of our parent(s)
      ** boundaries.  If so then we must restrict the exposed dimensions.
      */

      // THIS ROUTINE SUCKS - NEED TO ANALYSE ON RIGHT AND BOTTOM PATHS.

      ParentID  = list[index].ParentID;
      parentindex = index;
      while (ParentID) {
         while ((list[parentindex].ObjectID != ParentID) AND (parentindex > 0)) parentindex--;

         if (list[parentindex].BitmapID != list[index].BitmapID) break;

         ParentID     = list[parentindex].ParentID;
         parentx      = list[parentindex].AbsX;
         parenty      = list[parentindex].AbsY;
         parentwidth  = list[parentindex].Width;
         parentheight = list[parentindex].Height;

         /*** Check if the exposed area is out of bounds of the parent ***/

         if (((absx+XCoord+Width) <= parentx) OR ((absy+YCoord+Height) <= parenty) OR
             (absx+XCoord >= (parentx + parentwidth)) OR (absy+YCoord >= (parenty + parentheight))) {
            ReleaseRenderList(ARF_READ);
            return(ERR_Okay|ERF_Notified);
         }

         /*** Correct the height if the render object is located above or below the parent ***/

         if (absy+YCoord < parenty) {
            Height = (absy + Height) - parenty;
            YCoord = parenty - absy;
         }

         if (absy+YCoord+Height > parenty+parentheight) {
            Height = Height - (absy+YCoord+Height - (parenty+parentheight));
         }

         /*** Correct the width if the render object is located to the left or right of the parent ***/

         if (absx+XCoord < parentx) {
            Width  = (absx + Width) - parentx;
            XCoord = parentx - absx;
         }

         if (absx+XCoord+Width > parentx+parentwidth) {
            Width = Width - (absx+XCoord+Width - (parentx+parentwidth));
         }
      }

      /* Build an internal renderlist so that we can release the official renderlist for other tasks.
      ** This is important because some tasks can spend a lot of time in the drawing process and we
      ** shouldn't lock out the renderlist for very long.
      */

      for (i=0; list[i].ObjectID; i++);

      struct RenderList templist[i+1];

      CopyMemory(list, templist, sizeof(struct RenderList) * (i+1));
      ReleaseRenderList(ARF_READ);

      if (AccessObject(Self->BufferID, 5000, &bitmap) IS ERR_Okay) {

         /* Check if there has been a change in the video bit depth.  If so,
         ** regenerate the bitmap with a matching depth.
         */

         if ((GetScreenInfo(Self->ScreenID, &info, sizeof(info)) IS ERR_Okay) AND (info.BitsPerPixel != bitmap->BitsPerPixel)) {
            DPrintF("Draw:","[%d] Updating buffer bitmap %dx%dx%d to match new display depth of %dbpp.", bitmap->Head.UniqueID, bitmap->Width, bitmap->Height, bitmap->BitsPerPixel, info.BitsPerPixel);
            resize.Width  = bitmap->Width;
            resize.Height = bitmap->Height;
            resize.Depth  = info.BitsPerPixel;
            Action(AC_Resize, bitmap, &resize);

            Self->prvLineWidth     = bitmap->LineWidth;
            Self->prvBytesPerPixel = bitmap->BytesPerPixel;
            Self->prvBitsPerPixel  = bitmap->BitsPerPixel;
            Self->prvDataMID       = bitmap->DataMID;
            UpdateRenderList(Self);
         }

         CallbackRegion(Self, templist, index, absx+XCoord, absy+YCoord, absx+XCoord+Width, absy+YCoord+Height, bitmap);

         ReleaseObject(bitmap);

         return(ERR_Okay);
      }
      else return(SendError(ERH_Function, ERR_AccessObject));
   }
   else return(SendError(ERH_Function, ERR_AccessMemory));
}

static void CallbackRegion(struct Render *Self, struct RenderList *list, WORD Index,
                         WORD AbsX, WORD AbsY, WORD Right, WORD Bottom, struct Bitmap *DestBitmap)
{
   struct ClipRectangle abs;
   WORD i, j, level, listx, listy, listright, listbottom;

   if (Self->Flags & (RNF_REGION|RNF_INVISIBLE)) return;

   abs.left   = AbsX;
   abs.top    = AbsY;
   abs.right  = Right;
   abs.bottom = Bottom;

   /*** Make sure that the exposed clip area is restricted to the area of our drawable ***/

   if (abs.left   < list[Index].Left)   abs.left   = list[Index].Left;
   if (abs.top    < list[Index].Top)    abs.top    = list[Index].Top;
   if (abs.right  > list[Index].Right)  abs.right  = list[Index].Right;
   if (abs.bottom > list[Index].Bottom) abs.bottom = list[Index].Bottom;

   level = list[Index].Level + 1;   /* The +1 is used to include children contained in the render object */

   for (i=Index+1; list[i].Level > 1; i++) {
      if (list[i].Level < level) level = list[i].Level;

      /*** If the listed object obscures our render area, analyse the region around it ***/

      if (list[i].Level <= level) {
         /* If we have a bitmap buffer and the underlying child region also has its own bitmap,
         ** we have to ignore it in order for our graphics buffer to be correct when
         ** exposes are made.
         */

         if (list[i].BitmapID != Self->BufferID) continue;

         if ((!(list[i].Flags & RNF_VISIBLE)) OR (list[i].Flags & RNF_REGION)) {
            /* Regions are completely ignored because it is impossible
            ** for them to contain true render layers.
            */
            j = list[i].Level;
            while (list[i+1].Level > j) i++;
            continue;
         }

         /*** Check for an intersection and respond to it ***/

         listx = list[i].AbsX;
         listy = list[i].AbsY;
         listright = list[i].Right;
         listbottom = list[i].Bottom;

         if ((listx < Right) AND (listy < Bottom) AND (listright > AbsX) AND (listbottom > AbsY)) {

            if (list[i].Flags & RNF_CURSOR) {
               /* The pointer cursor is ignored */
               return;
            }
            else if (list[i].Flags & RNF_INVISIBLE) {
               /* If the render object is see-through then we will ignore its bounds, but legally
               ** it can also contain child render objects that are solid.  For that reason,
               ** we have to 'go inside' to check for solid children and draw around them.
               */

               CallbackRegion(Self, list, i, AbsX, AbsY, Right, Bottom, DestBitmap);

               return;
            }

            if (listx <= AbsX) listx = AbsX;
            else CallbackRegion(Self, list, Index, AbsX, AbsY, listx, Bottom, DestBitmap); /* left */

            if (listright >= Right) listright = Right;
            else CallbackRegion(Self, list, Index, listright, AbsY, Right, Bottom, DestBitmap); /* right */

            if (listy <= AbsY) listy = AbsY;
            else CallbackRegion(Self, list, Index, listx, AbsY, listright, listy, DestBitmap); /* top */

            if (listbottom < Bottom) CallbackRegion(Self, list, Index, listx, listbottom, listright, Bottom, DestBitmap); /* bottom */

            return;
         }
      }
   }

   /* If we have been called recursively due to the presence of volatile regions (see above),
   ** our Index field will be deeper than expected and we need to resync it before we
   ** do any drawing.
   */

   if (list[Index].ObjectID != Self->Head.UniqueID) {
      while ((Index > 0) AND (list[Index].ObjectID != Self->Head.UniqueID)) Index--;
   }

   /*** Prepare the buffer so that it matches the exposed area ***/

   if (Self->prvBitmapOwnerID != Self->Head.UniqueID) {
      for (i=Index; (i > 0) AND (list[i].ObjectID != Self->prvBitmapOwnerID); i--);
      DestBitmap->XOffset = list[Index].AbsX - list[i].AbsX; /* Offset is relative to the bitmap owner */
      DestBitmap->YOffset = list[Index].AbsY - list[i].AbsY;

   }
   else {
      /*** Set the clipping so that we only draw the area that has been exposed ***/
      DestBitmap->XOffset = 0;
      DestBitmap->YOffset = 0;
   }

   DestBitmap->ClipLeft   = AbsX - list[Index].AbsX;
   DestBitmap->ClipTop    = AbsY - list[Index].AbsY;
   DestBitmap->ClipRight  = Right - list[Index].AbsX;
   DestBitmap->ClipBottom = Bottom - list[Index].AbsY;

   // THIS SHOULD NOT BE NEEDED - but occasionally it detects render problems (bugs in other areas of the render code?)

   if (((DestBitmap->XOffset + DestBitmap->ClipLeft) < 0) OR ((DestBitmap->YOffset + DestBitmap->ClipTop) < 0) OR
       ((DestBitmap->XOffset + DestBitmap->ClipRight) > DestBitmap->Width) OR ((DestBitmap->YOffset + DestBitmap->ClipBottom) > DestBitmap->Height)) {
      DPrintF("@CallbackRegion()","Invalid coordinates detected (outside of the drawable area).");
      if ((DestBitmap->XOffset + DestBitmap->ClipLeft) < 0) DestBitmap->ClipLeft = 0;
      if ((DestBitmap->YOffset + DestBitmap->ClipTop) < 0)  DestBitmap->ClipTop = 0;
      DestBitmap->ClipRight = DestBitmap->Width - DestBitmap->XOffset;
      DestBitmap->ClipBottom = DestBitmap->Height - DestBitmap->YOffset;
   }

   /*** Call callbacks ***/

   if (Self->prvCallbackCount <= ARRAYSIZE(Self->prvCallbackCache)) {
      /*** Ensure that the prvCallback address is valid ***/
      Self->prvCallback = Self->prvCallbackCache;
   }

   for (i=0; i < Self->prvCallbackCount; i++) {
      void (*routine)(LONG, struct Render *, struct Bitmap *);
      routine = Self->prvCallback[i].Routine;
      routine(Self->prvCallback[i].Data, Self, DestBitmap);
   }
}
