/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Timer
Version:   1.0
Status:    Beta
Date:      November 2001
ID:        ID_TIMER
Category:  System
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2001.  All rights reserved.
Short:     Provides an activation service at regularly timed intervals..
Keywords:  timer, time, minute, second, wait, stop, timing, interrupt, signal

###DESCRIPTION###
<p>The Timer class provides developers with a simple mechanism for triggering
user defined actions at specific intervals or dates.  Once a timer object
has been set up, it will send Activate messages to its children whenever
a 'trigger' occurs.  Depending on your requirements, a trigger can be set
to occur at a given <a href="#Delay">interval</a> or a specific time on
a certain <a href="#Day">day</a>.</p>

<p>The Timer class is accurate to 1/1000th of a second, which makes it useful
for producing code that needs to run at a high frequency.  There are no
restrictions on the type of objects that can be placed in a Timer object;
but they must be able to support the Activate action in order to respond to
time signals.</p>

<p>If you need to reset a timer object so that its delay counter restarts
from zero, use the Reset action.  If you need a timer object to automatically
reset itself in response to user activity, refer to the options available
under the <a href="#Flags">Flags</a> field.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_TIMER 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_TIMER, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2001.  All rights reserved.",
   "November 2001",
   "Timer"
};

static struct KernelBase *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static OBJECTPTR TimerClass = 0;
static OBJECTPTR modStrings = 0;

#define TF_RESETKEYPRESS 0x00000001
#define TF_RESETMOUSE    0x00000002
#define TF_RESETIO       (TF_RESETKEYPRESS|TF_RESETMOUSE)

struct FieldDef TimerFlags[] = {
   { "RESETKEYPRESS", TF_RESETKEYPRESS }, { "RESETMOUSE", TF_RESETMOUSE },
   { "RESETIO",       TF_RESETIO },
   { NULL, NULL }
};

/*****************************************************************************
** Timer class.
*/

#define MAX_TIMERS 10

struct Timer {
   OBJECT_HEADER
   LONG   Flags;     /* Special flags */
   LONG   Count;     /* Number of times that activation has occurred */
   DOUBLE Delay;     /* Delay between each activation, in seconds */
   LONG   Limit;     /* Limits the total amount of times that activation can occur */

   PRIVATE_FIELDS
   OBJECTID KeyboardID;
   OBJECTID MouseID;
   WORD X, Y;
};

static ERROR GET_Delay(struct Timer *, DOUBLE *);

static ERROR SET_Delay(struct Timer *, DOUBLE *);

static struct FieldArray TimerFields[] = {
   { "Flags",    0, FDF_LONGFLAGS|FDF_RI, (LONG)&TimerFlags, NULL, NULL },
   { "Count",    0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Delay",    0, FDF_DOUBLE|FDF_RW,    0, NULL, NULL },
   { "Limit",    0, FDF_LONG|FDF_RW,      0, NULL, NULL },

   /*** Virtual fields ***/
   { "Interval", 0, FDF_DOUBLE|FDF_RI,    0, GET_Delay, SET_Delay },
   END_FIELD
};

static ERROR TIMER_ActionNotify(struct Timer *, struct acActionNotify *);
static ERROR TIMER_DataChannel(struct Timer *, struct acDataChannel *);
static ERROR TIMER_Free(struct Timer *, APTR);
static ERROR TIMER_Init(struct Timer *, APTR);
static ERROR TIMER_Reset(struct Timer *, APTR);
static ERROR TIMER_Timer(struct Timer *, struct acTimer *);

static struct ActionArray TimerActions[] = {
   { AC_ActionNotify, TIMER_ActionNotify },
   { AC_DataChannel,  TIMER_DataChannel },
   { AC_Free,         TIMER_Free },
   { AC_Init,         TIMER_Init },
   { AC_Reset,        TIMER_Reset },
   { AC_Timer,        TIMER_Timer },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&TimerClass, NULL,
      FID_BaseClassID|TLONG, ID_TIMER,
      FID_Version|TFLOAT,    VER_TIMER,
      FID_Name|TSTRING,      "Timer",
      FID_Category|TLONG,    CCF_SYSTEM,
      FID_Actions|TPTR,      TimerActions,
      FID_Fields|TPTR,       TimerFields,
      FID_Size|TLONG,        sizeof(struct Timer),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (modStrings) { acFree(modStrings); modStrings = NULL; }
   if (TimerClass) { acFree(TimerClass); TimerClass = NULL; }
   return(ERR_Okay);
}

/******************************************************************************
** Timer: ActionNotify
*/

static ERROR TIMER_ActionNotify(struct Timer *Self, struct acActionNotify *NotifyArgs)
{
   if (NotifyArgs->ActionID IS AC_MoveToPoint) {
      struct acMoveToPoint *moveto;
      WORD xc, yc;

      moveto = (struct acMoveToPoint *)NotifyArgs->Args;

      if (!((Self->X IS -1) AND (Self->Y IS -1))) {
         xc = moveto->XCoord - Self->X;
         yc = moveto->YCoord - Self->Y;
         if ((ABS(xc) > 2) OR (ABS(yc) > 2)) acReset(Self);
      }

      Self->X = moveto->XCoord;
      Self->Y = moveto->YCoord;
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Timer: DataChannel
*/

static ERROR TIMER_DataChannel(struct Timer *Self, struct acDataChannel *Args)
{
   /* When a Timer object receives information through its data channel,
   ** this causes the Timer to reset its interval.
   */

   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if (Args->DataType IS DATA_KEYPRESS) {
      acReset(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Timer: Free
*/

static ERROR TIMER_Free(struct Timer *Self, APTR Void)
{
   OBJECTPTR object;

   if (Self->KeyboardID) {
      if (AccessObject(Self->KeyboardID, 3000, &object) IS ERR_Okay) {
         UnsubscribeChannel(object, Self->Head.UniqueID);
         ReleaseObject(object);
      }
   }

   if (Self->MouseID) {
      if (AccessObject(Self->MouseID, 3000, &object) IS ERR_Okay) {
         UnsubscribeChannel(object, Self->Head.UniqueID);
         ReleaseObject(object);
      }
   }

   UnsubscribeTimer(Self->Head.UniqueID);
   return(ERR_Okay);
}

/*****************************************************************************
** Timer: Init
*/

static ERROR TIMER_Init(struct Timer *Self, APTR Void)
{
   OBJECTPTR keyboard, mouse;
   LONG count;

   if (Self->Delay < 0.01) {
      DPrintF("@Init:","[Timer:%d] The Delay field is not set.", Self->Head.UniqueID);
      return(ERR_FieldNotSet);
   }

   if (Self->Flags & TF_RESETKEYPRESS) {
      /*** Listen to the SystemKeyboard object to detect key presses ***/

      if (!Self->KeyboardID) {
         if (FastFindObject("SystemKeyboard", NULL, &Self->KeyboardID, 1, &count) IS ERR_Okay) {
            if (AccessObject(Self->KeyboardID, 3000, &keyboard) IS ERR_Okay) {
               SubscribeChannel(keyboard, Self);
               ReleaseObject(keyboard);
            }
         }
         else DPrintF("@Init:","[Timer:%d] Failed to find a SystemKeyboard object.", Self->Head.UniqueID);
      }
   }

   if (Self->Flags & TF_RESETMOUSE) {
      /*** Listen to the mouse pointer to detect user movement ***/

      if (!Self->MouseID) {
         if (FastFindObject("SystemPointer", NULL, &Self->MouseID, 1, &count) IS ERR_Okay) {
            if (AccessObject(Self->MouseID, 3000, &mouse) IS ERR_Okay) {
               SubscribeActionTags(mouse, Self, AC_MoveToPoint, TAGEND);
               ReleaseObject(mouse);
            }
         }
         else DPrintF("@Init:","[Timer:%d] Failed to find a SystemPointer object.", Self->Head.UniqueID);
      }
   }

   /*** Initialise the Timer ***/

   Self->X = -1;
   Self->Y = -1;

   if (SubscribeTimer(Self->Head.UniqueID, Self->Delay * 1000) IS ERR_Okay) {
      return(ERR_Okay);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###ACTION###
Name:  Reset
Short: Resets a timer object.
###END###

*****************************************************************************/

static ERROR TIMER_Reset(struct Timer *Self, APTR Void)
{
   /*** Re-subscribing to the system timer will reset the tick count to 0 ***/
   SubscribeTimer(Self->Head.UniqueID, Self->Delay * 1000);
   return(ERR_Okay);
}

/*****************************************************************************
** Timer: Timer
*/

static ERROR TIMER_Timer(struct Timer *Self, struct acTimer *Args)
{
   #define MAX_OBJECTS 20
   struct ChildEntry list[MAX_OBJECTS];
   LONG count, i;

   Self->Count++; /* Increment trigger count */

   /*** Activate children ***/

   count = MAX_OBJECTS;
   if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {
      for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
   }

   /*** Check if the limit has been reached ***/

   if ((Self->Limit > 0) AND (Self->Count >= Self->Limit)) {
      acFree(Self);
   }

   Self->X = -1;
   Self->Y = -1;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Count
Short:  Indicates the total amount of times that a Timer object has been triggered.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>The Count field is incremented each time that a Timer object is triggered.
Reading the Count field will give you an accurate indication as to how many
times activation has occurred.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Date
Short:  Determines a date and/or time at which a Timer object should be triggered.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The Date allows you to define a trigger that occurs at a specific date, or a
set time each day.  The allowed formats for the date string are as follows:</p>

<ul>
  <li>Day/Month/Year Hour:Min:Sec</li>
  <li>Hour:Min:Sec</li>
</ul>

<p>Here are some examples:</li>

<ul>
  <li>24/10/2001 18:05:15</li>
  <li>8/5/2002 7:59</li>
  <li>8:00</li>
  <li>23:12:52</li>
</ul>

<p>If you set a time value without a specific date, then activation will occur
at that time, every day of the week.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    Delay
Synonym: Interval
Short:   The number of seconds to wait between each activation.
Type:    DOUBLE
Status:  Read/Write

###DESCRIPTION###
<p>If a timer object will be activated at regular intervals, you need to set
the Delay field to determine the exact number of seconds between each
activation.  This field is floating point based, which allows you to specify
an interval with an accuracy down to 1/1000th of a second.</p>

###END###

*****************************************************************************/

static ERROR GET_Delay(struct Timer *Self, DOUBLE *Value)
{
   *Value = Self->Delay;
   return(ERR_Okay);
}

static ERROR SET_Delay(struct Timer *Self, DOUBLE *Value)
{
   Self->Delay = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Limit
Short:  Places a limit on the total amount of times that activation can occur.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The Limit field alllows you to limit the total number of times that a Timer
object can be triggered.  For instance, if you would like to set an alarm that
goes off in 3 minutes without a trigger occurring at 6, 9, 12 minutes or more,
then you would set the Limit to a value of 1.  Once the limit is met, the
Timer object will automatically free itself from the system to prevent it
from wasting system resources.</p>

<p>If the Limit is set to zero or less, the feature is turned off.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Special flags.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>You can specify flags in this field that will affect the behaviour of a
Timer object.  Valid flags are:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">RESETKEYPRESS&nbsp;&nbsp;</td><td>This flag tells a Timer object to reset its counter whenever a keypress is received.</td></tr>
<tr><td width="10%">RESETMOUSE</td><td>This flag tells a Timer object to reset its counter whenever the user moves the mouse.</td></tr>
<tr><td width="10%">RESETIO</td><td>If this flag is set, the timing counter will reset itself when any kind of user activity is detected (keypresses, mouse movement, joyport movement etc).</td></tr>
</table>

###END###

*****************************************************************************/
